"""Build Process
"""
import sys
import os
import re
import copy
import json
import yaml
import openapi_spec_validator
import jsonpath_ng
import inspect

try:
    from typing import Union, Dict, Literal
except ImportError:
    from typing_extensions import Literal


class AutoFieldUid(object):
    def __init__(self):
        self._field_uid = 0

    @property
    def uid(self):
        self._field_uid += 1
        return self._field_uid


class Bundler(object):
    """Bundles OpenAPI yaml files into a single API document

    Notes
    -----
    - Bundles multiple OpenAPI v3.x yaml files into a single file.
    - Does not inline schemas indicated by ref keywords but normalizes them by
      removing the file path.
    - Validates the bundled file

    Args
    ----
        output_dir (str): The directory where files will be output
        api_files (str): The top level api files
    """

    class description(str):
        pass

    @staticmethod
    def literal_representer(dumper, data):
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", data, style="|"
        )

    def __init__(self, api_files, output_dir="./", generate_version_api=False):
        self._parsers = {}
        self._api_files = api_files
        self._output_dir = os.path.abspath(output_dir)
        self._generate_version_api = generate_version_api
        if os.path.exists(self._output_dir) is False:
            os.makedirs(self._output_dir)
        self.__python = os.path.normpath(sys.executable)
        self._content = {}
        self._includes = {}
        self._include_objects = {}
        self._resolved = []
        self._errors = []
        yaml.add_representer(Bundler.description, Bundler.literal_representer)

    def _get_parser(self, pattern):
        if pattern not in self._parsers:
            parser = jsonpath_ng.parse(pattern)
            self._parsers[pattern] = parser
        else:
            parser = self._parsers[pattern]
        return parser

    @property
    def openapi_filepath(self):
        return self._output_filename

    def bundle(self):
        self._errors = []
        self._output_filename = os.path.join(self._output_dir, "openapi.yaml")
        self._json_filename = os.path.join(self._output_dir, "openapi.json")
        self._content = {}
        self._includes = {}
        self._include_objects = {}
        self._resolved = []
        for api_filename in self._api_files:
            api_filename = os.path.normpath(os.path.abspath(api_filename))
            self._base_dir = os.path.dirname(api_filename)
            self._api_filename = os.path.basename(api_filename)
            self._read_file(self._base_dir, self._api_filename)

        self._resolve_x_include()
        self._resolve_x_pattern("x-field-pattern")
        self._resolve_x_constraint()
        self._resolve_x_status()
        self._remove_x_include()
        # TODO: restore behavior
        # self._resolve_x_unique()
        self._resolve_license()
        self._resolve_x_enum(self._content)
        self._generate_version_api_spec(self._content)
        self._validate_field_uid()
        self._validate_response_uid()
        self._validate_errors()
        self._validate_required_responses()
        self._resolve_strings(self._content)
        self._resolve_keys(self._content)
        self._api_version = self._content["info"]["version"]
        with open(self._output_filename, "w") as fp:
            yaml.dump(
                self._content,
                fp,
                indent=2,
                allow_unicode=True,
                line_break="\n",
                sort_keys=False,
            )
        with open(self._json_filename, "w") as fp:
            fp.write(json.dumps(self._content, indent=4))
        self._validate_file()

    def get_api_version(self):
        return self._api_version

    def _validate_errors(self):
        if len(self._errors) > 0:
            raise TypeError("\n".join(self._errors))

    def _inject_enum(self, property_name, property_object, schema_name):
        # if "enum" in property_object.keys():
        #     self._errors.append(
        #         "Please modify enum with x-enum within %s:%s"
        #         % (schema_name, property_name)
        #     )
        #     return
        if "x-enum" in property_object.keys():
            property_object["enum"] = list(property_object["x-enum"].keys())

    def _resolve_x_enum(self, content):
        for schema_name, schema_object in content["components"][
            "schemas"
        ].items():
            if "properties" not in schema_object:
                continue
            for property_name, property_object in schema_object[
                "properties"
            ].items():
                if "type" not in property_object.keys():
                    continue
                if property_object["type"] == "string":
                    self._inject_enum(
                        property_name, property_object, schema_name
                    )
                if (
                    property_object["type"] == "array"
                    and "items" in property_object
                ):
                    self._inject_enum(
                        property_name, property_object["items"], schema_name
                    )

    def _generate_version_api_spec(self, content):
        if not self._generate_version_api:
            return

        print("Generating version API ...")
        schema_name = "Version"
        schema_ref = "#/components/schemas/{}".format(schema_name)
        api_path = "/capabilities/version"

        if schema_name in content["components"]["schemas"]:
            schema = content["components"]["schemas"][schema_name]
            for prop in ["api_spec_version", "sdk_version", "app_version"]:
                if prop not in schema["properties"]:
                    raise AssertionError(
                        "Could not generate version schema: Version is missing property {}".format(
                            prop
                        )
                    )
                if schema["properties"][prop]["type"] != "string":
                    raise AssertionError(
                        "Could not generate version schema: Version property {} MUST be of type string".format(
                            prop
                        )
                    )
        else:
            content["components"]["schemas"][schema_name] = {
                "description": "Version details",
                "type": "object",
                "properties": {
                    "api_spec_version": {
                        "description": "Version of API specification",
                        "type": "string",
                        "default": "",
                        "x-field-uid": 1,
                    },
                    "sdk_version": {
                        "description": "Version of SDK generated from API specification",
                        "type": "string",
                        "default": "",
                        "x-field-uid": 2,
                    },
                    "app_version": {
                        "description": "Version of application consuming or serving the API",
                        "type": "string",
                        "default": "",
                        "x-field-uid": 3,
                    },
                },
            }

        if api_path in content["paths"]:
            assert (
                content["paths"][api_path]["get"]["responses"]["200"][
                    "content"
                ]["application/json"]["schema"]["$ref"]
                == schema_ref
            ), "{} MUST have a 200 GET response {}".format(
                api_path, schema_ref
            )
        else:
            err_responses = {}
            first_path = list(content["paths"].keys())[0]
            first_method = list(content["paths"][first_path].keys())[0]
            for code, res in content["paths"][first_path][first_method][
                "responses"
            ].items():
                code_str = str(code)
                if (
                    not code_str.startswith("1")
                    and not code_str.startswith("2")
                    and not code_str.startswith("3")
                ):
                    err_responses[code] = res

            content["paths"][api_path] = {
                "get": {
                    "tags": ["Capabilities"],
                    "operationId": "get_version",
                    "responses": {
                        "200": {
                            "description": "Version details from API server",
                            "content": {
                                "application/json": {
                                    "schema": {"$ref": schema_ref}
                                }
                            },
                            "x-field-uid": 1,
                        },
                    },
                },
            }

            for code, res in err_responses.items():
                content["paths"][api_path]["get"]["responses"][
                    code
                ] = copy.deepcopy(res)

    def _check_duplicate_uid(self, fields_uid, name):
        dup_values = set([x for x in fields_uid if fields_uid.count(x) > 1])
        if len(dup_values) > 0:
            self._errors.append(
                "%s contain duplicate %s x-field-uid. x-field-uid should be unique."
                % (name, list(dup_values))
            )

    def _check_range_uid(self, fields_uid, name):
        if fields_uid < 0 or fields_uid > 536870911:
            self._errors.append(
                "x-field-uid %s of %s not in range (1 to 2^29)"
                % (fields_uid, name)
            )

    def _validate_xenum_field_uid(self, property_name, property_object):
        reserved_field_uids = []
        enums = property_object["x-enum"]
        if "x-reserved-field-uids" in property_object:
            reserved_field_uids = property_object["x-reserved-field-uids"]
        field_uids = []
        for key, value in enums.items():
            if "x-field-uid" not in value:
                self._errors.append("x-field-uid is missing in %s" % key)
                continue
            field_uid = value["x-field-uid"]
            field_uids.append(field_uid)
            if field_uid in reserved_field_uids:
                self._errors.append(
                    "x-field-uid %s within enum %s:%s conflict with x-reserved-field-uids"
                    % (field_uid, property_name, key)
                )
            self._check_range_uid(
                field_uid, "{}:{}".format(property_name, key)
            )
        self._check_duplicate_uid(field_uids, property_name)

    def _validate_field_uid(self):
        for schema_name, schema_object in self._content["components"][
            "schemas"
        ].items():
            if "properties" not in schema_object:
                continue
            field_uids = []
            reserved_field_uids = []
            if "x-reserved-field-uids" in schema_object:
                reserved_field_uids = schema_object["x-reserved-field-uids"]
            for property_name, property_object in schema_object[
                "properties"
            ].items():
                field_uid = property_object.get("x-field-uid")
                if field_uid is None:
                    self._errors.append(
                        "x-field-uid is missing in %s:%s"
                        % (schema_name, property_name)
                    )
                    continue
                self._check_range_uid(
                    field_uid, "{}:{}".format(schema_name, property_name)
                )
                field_uids.append(field_uid)
                if field_uid in reserved_field_uids:
                    self._errors.append(
                        "x-field-uid %s of %s:%s should not conflict with x-reserved-field-uids"
                        % (field_uid, schema_name, property_name)
                    )
                if "x-enum" in property_object:
                    self._validate_xenum_field_uid(
                        property_name, property_object
                    )
            self._check_duplicate_uid(field_uids, schema_name)

    def _validate_response_uid(self):
        for path_key, path_object in self._content["paths"].items():
            for path_item_key, path_item_object in path_object.items():
                field_uids = []
                reserved_field_uids = []
                if "x-reserved-field-uids" in path_item_object:
                    reserved_field_uids = path_item_object[
                        "x-reserved-field-uids"
                    ]
                for response in self._get_parser("$..responses").find(
                    path_item_object
                ):
                    for code, code_schema in response.value.items():
                        field_uid = code_schema.get("x-field-uid")
                        common_name = "{}:{}:{}".format(
                            path_key, path_item_key, code
                        )
                        if field_uid is None:
                            self._errors.append(
                                "x-field-uid is missing in %s response"
                                % common_name
                            )
                            continue
                        field_uids.append(field_uid)
                        self._check_range_uid(field_uid, common_name)
                        if field_uid in reserved_field_uids:
                            self._errors.append(
                                "x-field-uid %s of %s should not conflict with x-reserved-field-uids"
                                % (field_uid, common_name)
                            )
                    self._check_duplicate_uid(
                        field_uids, "{}:{}".format(path_key, path_item_key)
                    )

    def _validate_required_responses(self):
        """Ensure all paths include a 400 and 500 response.

        Print every path that does not include a 400 or 500 response.

        Returns
        -------
        Exception: one or more paths is missing a 400 or 500 response
        None: all paths have a 400 and 500 response
        """
        responses = self._get_parser("$..paths..responses").find(self._content)
        required_error_codes = ["400", "500"]
        missing_paths = ""
        for response in responses:
            missing = set(required_error_codes).difference(
                set(response.value.keys())
            )
            if len(missing):
                error_message = "{}: is missing the following required responses: {}".format(
                    response.full_path,
                    missing,
                )
                print(error_message)
                missing_paths += "{}\n".format(error_message)
        if len(missing_paths) > 0:
            raise Exception(missing_paths)
        return None

    def _validate_file(self):
        print("validating {}...".format(self._output_filename))
        with open(self._output_filename) as fid:
            yobject = yaml.safe_load(fid)
            openapi_spec_validator.validate_v3_spec(yobject)
        print("validating complete")

    def _read_file(self, base_dir, filename):
        filename = os.path.join(base_dir, filename)
        filename = os.path.abspath(os.path.normpath(filename))
        base_dir = os.path.dirname(filename)
        with open(filename) as fid:
            yobject = yaml.safe_load(fid)
        self._process_yaml_object(base_dir, yobject)

    def _process_yaml_object(self, base_dir, yobject):
        for key, value in yobject.items():
            if (
                key in ["openapi", "info", "servers"]
                and key not in self._content.keys()
            ):
                self._content[key] = value
            elif key in ["paths"]:
                if key not in self._content.keys():
                    self._content[key] = {}
                for sub_key in value.keys():
                    self._content[key][sub_key] = value[sub_key]
            elif key == "components":
                if key not in self._content.keys():
                    self._content[key] = {"responses": {}, "schemas": {}}
                self._validate_names("^[+a-zA-Z0-9_]+$", "schemas", value)
                self._validate_names("^[+a-zA-Z0-9_]+$", "responses", value)
                self._check_nested_components(value)
        self._resolve_refs(base_dir, yobject)

    def _check_upper_case(self, value):

        for c in value:
            if c.isupper():
                return True

        return False

    def _validate_names(self, regex, components_key, components):
        if components_key not in components:
            return
        objects = components[components_key]
        for key, value in objects.items():
            if "properties" in objects[key]:
                for name in objects[key]["properties"]:
                    if re.match(regex, name) is None:
                        raise NameError(
                            "%s property name `%s` contains invalid characters"
                            % (key, name)
                        )
                    if self._check_upper_case(name):
                        raise NameError(
                            "*** Property name '{}' is invalid. Only lower case letters separated with an underscore is allowed.***".format(
                                name
                            )
                        )

            self._content["components"][components_key][key] = value

    def _check_nested_components(self, components):
        objects = components["schemas"]
        errors = []
        for component_name, component_value in objects.items():
            if "properties" in component_value:
                component_properties = component_value["properties"]
                for property_name in component_properties:
                    property_obj = component_properties[property_name]
                    if "type" in property_obj:
                        proprty_type = property_obj["type"]
                        if proprty_type == "object":
                            errors.append(
                                "\n*** Unsupported. Property '{}'.'{}' is a nested component ***".format(
                                    component_name,
                                    property_name,
                                )
                            )
        if len(errors) > 0:
            raise TypeError("".join(errors))

    def _resolve_refs(self, base_dir, yobject):
        """Resolving references is relative to the current file location"""
        if isinstance(yobject, dict):
            for key, value in yobject.items():
                if key == "$ref" and value.startswith("#") is False:
                    refs = value.split("#")
                    if refs[1] not in self._resolved:
                        self._resolved.append(refs[1])
                        print("resolving %s" % value)
                        self._read_file(base_dir, refs[0])
                    yobject[key] = "#%s" % refs[1]
                elif isinstance(value, str) and "x-inline" in value:
                    refs = value.split("#")
                    print("inlining %s" % value)
                    inline = self._get_inline_ref(base_dir, refs[0], refs[1])
                    yobject[key] = inline
                elif key == "x-include":
                    if not isinstance(value, str):
                        self._errors.append(
                            "%s of x-include shall be a path of any property or response"
                            % value
                        )
                        continue
                    if value not in self._includes:
                        file_name, include_path = value.split("#")
                        if "properties" in include_path:
                            obj_path, path_name = include_path.split(
                                "properties"
                            )
                            include_ref = "{}#{}".format(
                                file_name, obj_path[:-1]
                            )
                            path_name = "properties{}".format(path_name)
                        elif "responses" in include_path:
                            obj_path, path_name = include_path.split(
                                "responses"
                            )
                            path_list = path_name[1:].split("/")
                            include_ref = "{}#{}responses/{}".format(
                                file_name, obj_path, path_list[0]
                            )
                            path_name = "/".join(path_list[1:])
                        else:
                            self._errors.append(
                                "x-include should ref some properties/ responses"
                            )
                            continue
                        if include_ref in self._include_objects:
                            include_object = self._include_objects[include_ref]
                        else:
                            include_object = self._get_schema_object(
                                base_dir, include_ref
                            )
                            self._include_objects[include_ref] = include_object
                        field_object = self._get_field_object(
                            include_object, path_name
                        )
                        self._includes[value] = field_object
                        self._resolve_refs(base_dir, field_object)
                else:
                    self._length_restriction(value)
                    self._required_restriction(key, value)
                    self._resolve_refs(base_dir, value)
        elif isinstance(yobject, list):
            for item in yobject:
                self._resolve_refs(base_dir, item)

    def _get_field_object(self, yobject, field_path):
        field_object = yobject
        for node in field_path.split("/"):
            tmp = field_object.get(node)
            if tmp is None and node.isdigit():
                tmp = field_object.get(int(node))
            field_object = tmp
        return field_object

    def _length_restriction(self, value):
        restricted_keys = {
            "length",
            "minimum",
            "maximum",
            "minLength",
            "maxLength",
        }
        if isinstance(value, dict):
            intersect_keys = restricted_keys.intersection(set(value.keys()))
            if (
                len(intersect_keys) > 0
                and "format" in value.keys()
                and value["format"] in ["ipv4", "ipv6", "mac"]
            ):
                stacks = inspect.stack()
                property = "{}/{}/{}".format(
                    stacks[3].frame.f_locals["key"]
                    if "key" in stacks[3].frame.f_locals
                    else "",
                    stacks[2].frame.f_locals["key"]
                    if "key" in stacks[2].frame.f_locals
                    else "",
                    stacks[1].frame.f_locals["key"],
                )
                self._errors.append(
                    "Property {property} should not contain {keys} with format {format}".format(
                        property=property,
                        keys=intersect_keys,
                        format=value["format"],
                    )
                )

    def _required_restriction(self, schema_name, value):
        expected_set = {"required", "properties"}
        if isinstance(value, dict) and expected_set.issubset(value.keys()):
            if isinstance(value["required"], list):
                for required in value["required"]:
                    if "default" in value["properties"][required].keys():
                        self._errors.append(
                            "Property {property} within schema {name} have "
                            "both required as well as default".format(
                                property=required, name=schema_name
                            )
                        )
                    if "x-status" in value["properties"][required].keys():
                        if (
                            value["properties"][required]["x-status"]["status"]
                            == "deprecated"
                        ):
                            raise Exception(
                                "Property {property} within schema {name} have "
                                "both required as well as deprecated status".format(
                                    property=required, name=schema_name
                                )
                            )

    def _resolve_x_pattern(self, pattern_extension):
        """Find all instances of pattern_extension in the openapi content
        and generate a #/components/schemas/... pattern schema object that is
        specific to the property hosting the pattern extension content.
        Replace the x-field-pattern schema with a $ref to the generated schema.
        """
        import jsonpath_ng

        for xpattern_path in self._get_parser(
            "$..{}".format(pattern_extension)
        ).find(self._content):
            print("generating %s..." % (str(xpattern_path.full_path)))
            object_name = xpattern_path.full_path.left.left.left.right.fields[
                0
            ]
            property_name = xpattern_path.full_path.left.right.fields[0]
            property_schema = jsonpath_ng.Parent().find(xpattern_path)[0].value
            xpattern = xpattern_path.value
            schema_name = "Pattern.{}.{}".format(
                "".join(
                    [
                        piece[0].upper() + piece[1:]
                        for piece in object_name.split("_")
                    ]
                ),
                "".join(
                    [
                        piece[0].upper() + piece[1:]
                        for piece in property_name.split("_")
                    ]
                ),
            )
            format = None
            type_name = xpattern["format"]
            if type_name in ["ipv4", "ipv6", "mac", "x-enum"]:
                format = type_name
                type_name = "string"
            description = "TBD"
            if "description" in xpattern:
                description = xpattern["description"]
            elif "description" in property_schema:
                description = property_schema["description"]

            if xpattern["format"] == "checksum":
                self._generate_checksum_schema(
                    xpattern, schema_name, description
                )
            else:
                self._generate_value_schema(
                    xpattern, schema_name, description, type_name, format
                )

            property_schema["$ref"] = "#/components/schemas/{}".format(
                schema_name
            )
            del property_schema[pattern_extension]

    def _generate_checksum_schema(self, xpattern, schema_name, description):
        """Generate a checksum schema object"""
        auto_field = AutoFieldUid()
        schema = {
            "description": description,
            "type": "object",
            "properties": {
                "choice": {
                    "description": "The type of checksum",
                    "type": "string",
                    "x-enum": {
                        "generated": {"x-field-uid": 1},
                        "custom": {"x-field-uid": 2},
                    },
                    "default": "generated",
                    "x-field-uid": auto_field.uid,
                },
                "generated": {
                    "description": "A system generated checksum value",
                    "type": "string",
                    "x-enum": {
                        "good": {"x-field-uid": 1},
                        "bad": {"x-field-uid": 2},
                    },
                    "default": "good",
                    "x-field-uid": auto_field.uid,
                },
                "custom": {
                    "description": "A custom checksum value",
                    "type": "integer",
                    "minimum": 0,
                    "maximum": 2 ** int(xpattern.get("length", 8)) - 1,
                    "x-field-uid": auto_field.uid,
                },
            },
        }
        self._content["components"]["schemas"][schema_name] = schema

    def _generate_value_schema(
        self, xpattern, schema_name, description, type_name, format
    ):
        auto_field = AutoFieldUid()
        xconstants = (
            xpattern["x-constants"] if "x-constants" in xpattern else None
        )
        schema = {
            "description": description,
            "type": "object",
            "properties": {
                "choice": {
                    "type": "string",
                    "x-enum": {
                        "value": {"x-field-uid": 2},
                        "values": {"x-field-uid": 3},
                    },
                    "default": "value",
                    "x-field-uid": auto_field.uid,
                },
                "value": {
                    "type": copy.deepcopy(type_name),
                    "x-field-uid": auto_field.uid,
                },
                "values": {
                    "type": "array",
                    "items": {"type": copy.deepcopy(type_name)},
                    "x-field-uid": auto_field.uid,
                },
            },
        }
        if xconstants is not None:
            schema["x-constants"] = copy.deepcopy(xconstants)
        if "features" in xpattern:
            if "auto" in xpattern["features"]:
                if "default" not in xpattern:
                    self._errors.append(
                        "default must be set for property {}, when auto feature is enabled".format(
                            schema_name
                        )
                    )
                schema["properties"]["choice"]["x-enum"]["auto"] = {
                    "x-field-uid": 1
                }
                schema["properties"]["choice"]["default"] = "auto"
                description = [
                    "The OTG implementation can provide a system generated",
                    "value for this property. If the OTG is unable to generate a value",
                    "the default value must be used.",
                ]
                schema["properties"]["auto"] = {
                    "description": "\n".join(description),
                    "type": copy.deepcopy(type_name),
                    "x-field-uid": auto_field.uid,
                }
                self._apply_common_x_field_pattern_properties(
                    schema["properties"]["auto"],
                    xpattern,
                    format,
                    property_name="auto",
                )
            if "metric_group" in xpattern["features"]:
                schema["properties"]["metric_group"] = {
                    "description": """A unique name is used to indicate to the system that the field may """
                    """extend the metric row key and create an aggregate metric row for """
                    """every unique value. """
                    """To have metric group columns appear in the flow metric rows the flow """
                    """metric request allows for the metric_group value to be specified """
                    """as part of the request.""",
                    "type": "string",
                    "x-field-uid": auto_field.uid,
                }
        if "enums" in xpattern:
            schema["properties"]["value"]["enum"] = copy.deepcopy(
                xpattern["enums"]
            )
            schema["properties"]["values"]["items"]["enum"] = copy.deepcopy(
                xpattern["enums"]
            )
        if xpattern["format"] in ["integer", "ipv4", "ipv6", "mac"]:
            counter_pattern_name = "{}.Counter".format(schema_name)
            schema["properties"]["choice"]["x-enum"]["increment"] = {
                "x-field-uid": 4
            }
            schema["properties"]["choice"]["x-enum"]["decrement"] = {
                "x-field-uid": 5
            }
            schema["properties"]["increment"] = {
                "$ref": "#/components/schemas/{}".format(counter_pattern_name)
            }
            schema["properties"]["increment"]["x-field-uid"] = auto_field.uid
            schema["properties"]["decrement"] = {
                "$ref": "#/components/schemas/{}".format(counter_pattern_name)
            }
            schema["properties"]["decrement"]["x-field-uid"] = auto_field.uid
            counter_auto_field = AutoFieldUid()
            counter_schema = {
                "description": "{} counter pattern".format(xpattern["format"]),
                "type": "object",
                "properties": {
                    "start": {
                        "type": type_name,
                        "x-field-uid": counter_auto_field.uid,
                    },
                    "step": {
                        "type": type_name,
                        "x-field-uid": counter_auto_field.uid,
                    },
                },
            }
            if "features" in xpattern and "count" in xpattern["features"]:
                counter_schema["properties"]["count"] = {
                    "type": "integer",
                    "default": 1,
                    "x-field-uid": counter_auto_field.uid,
                }
            self._apply_common_x_field_pattern_properties(
                counter_schema["properties"]["start"],
                xpattern,
                format,
                property_name="start",
            )
            self._apply_common_x_field_pattern_properties(
                counter_schema["properties"]["step"],
                xpattern,
                format,
                property_name="step",
            )
            if xconstants is not None:
                counter_schema["x-constants"] = copy.deepcopy(xconstants)
            self._content["components"]["schemas"][
                counter_pattern_name
            ] = counter_schema
        self._apply_common_x_field_pattern_properties(
            schema["properties"]["value"],
            xpattern,
            format,
            property_name="value",
        )
        self._apply_common_x_field_pattern_properties(
            schema["properties"]["values"],
            xpattern,
            format,
            property_name="values",
        )
        self._content["components"]["schemas"][schema_name] = schema

    def _apply_common_x_field_pattern_properties(
        self, schema, xpattern, format, property_name
    ):
        # type: (Dict, Dict, str, Union[Literal["start"], Literal["step"], Literal["value"], Literal["values"]])
        step_defaults = {
            "mac": "00:00:00:00:00:01",
            "ipv4": "0.0.0.1",
            "ipv6": "::1",
        }
        if "default" in xpattern:
            schema["default"] = xpattern["default"]
            if property_name == "step":
                if format in step_defaults:
                    schema["default"] = step_defaults[format]
                else:
                    schema["default"] = 1
            elif property_name == "values":
                schema["default"] = [schema["default"]]
        if format is not None:
            # TODO: fix this
            # if property_name == "values":
            #     schema["items"]["format"] = format
            # else:
            schema["format"] = format
        if "length" in xpattern:
            # TODO: fix this
            # if property_name == "values":
            #     schema["items"]["minimum"] = 0
            #     schema["items"]["maximum"] = 2 ** int(xpattern["length"]) - 1
            # else:
            schema["minimum"] = 0
            schema["maximum"] = 2 ** int(xpattern["length"]) - 1

    def _resolve_recursive_x_include(self, include_value):
        if "x-include" in include_value:
            include_schema_object = self._includes[include_value["x-include"]]
            if "x-include" in include_schema_object:
                self._resolve_recursive_x_include(include_schema_object)
                include_schema_object = self._includes[
                    include_value["x-include"]
                ]
            self._merge(copy.deepcopy(include_schema_object), include_value)

    def _resolve_x_include(self):
        """Find all instances of x-include in the openapi content
        and merge the x-include content into the parent object
        Remove the x-include and the included content
        """
        for include_key, include_value in self._includes.items():
            self._resolve_recursive_x_include(include_value)

        for xinclude in self._get_parser("$..x-include").find(self._content):
            parent_schema_object = jsonpath_ng.Parent().find(xinclude)[0].value
            xinclude_value = xinclude.value
            print("resolving %s..." % (str(xinclude_value)))
            if (
                len(self._includes) == 0
                or xinclude_value not in self._includes
            ):
                self._errors.append(
                    "x-include %s missing in internal object."
                    "x-include should define within properties"
                    % xinclude_value
                )
                continue
            include_schema_object = self._includes[xinclude_value]
            self._merge(
                copy.deepcopy(include_schema_object), parent_schema_object
            )
            del parent_schema_object["x-include"]

    def _remove_x_include(self):
        refs = self._get_parser('$.."$ref"').find(self._content)
        for item in set(self._include_objects):
            pieces = item.split("#/")[1].split("/")
            value = "#/{}".format("/".join(pieces))
            match = False
            for ref in refs:
                if ref.value == value:
                    match = True
                    break
            if match is False:
                content = self._content
                for piece in pieces[0:-1]:
                    content = content[piece]
                if pieces[-1] in content:
                    del content[pieces[-1]]

    def _resolve_x_status(self):
        """Find all instances of x-status in the openapi content
        and merge the x-status content into the parent object description
        """
        import jsonpath_ng

        valid_statuses = ["deprecated", "under_review"]
        for xstatus in self._get_parser("$..x-status").find(self._content):
            status = xstatus.value.get("status")

            if status is not None:
                status = status.replace("-", "_")

            if status not in valid_statuses:
                raise Exception(
                    "Invalid value for x-status.status={} provided; Valid values are {}".format(
                        status, valid_statuses
                    )
                )

            print("resolving {} ...".format(xstatus.full_path))
            parent = jsonpath_ng.Parent().find(xstatus)[0]
            parent_schema = parent.value

            info = xstatus.value.get("information")
            if info is None or len(info) == 0:
                info = xstatus.value["information"] = "Information TBD"
                print(
                    "[WARNING]: {}.x-status.information missing".format(
                        parent.full_path
                    )
                )

            if status == "deprecated":
                status_msg = "Deprecated: {}".format(info)
            elif status == "under_review":
                status_msg = "Under Review: {}".format(info)

            desc = parent_schema.get("description")
            if desc is None or len(desc) == 0:
                desc = "Description TBD"

            parent_schema["description"] = "{}\n\n{}".format(status_msg, desc)

    def _resolve_x_unique(self):
        """validate the x-unique field and make sure it is [global]"""
        for xunique in self._get_parser("$..x-unique").find(self._content):
            if xunique.value in ["global"]:
                continue
            raise Exception("x-unique can have only 'global'")

    def _resolve_x_constraint(self):
        """Find all instances of x-constraint in the openapi content
        and merge the x-constraint content into the parent object description
        """
        import jsonpath_ng

        for xconstraint in self._get_parser("$..x-constraint").find(
            self._content
        ):
            print("resolving %s..." % (str(xconstraint.full_path)))
            parent_schema_object = (
                jsonpath_ng.Parent().find(xconstraint)[0].value
            )
            if "description" not in parent_schema_object:
                parent_schema_object["description"] = "TBD"
            parent_schema_object["description"] += "\n\nx-constraint:\n"
            for constraint in xconstraint.value:
                parent_schema_object["description"] += "- {}\n".format(
                    constraint
                )

    def _merge(self, src, dst):
        """
        Recursively update a dict.
        Subdict's won't be overwritten but also updated.
        """
        for key, value in src.items():
            if key not in dst:
                dst[key] = value
            elif key == "x-field-uid":
                continue
            elif isinstance(value, list):
                for item in value:
                    if item not in dst[key]:
                        dst[key].append(item)
            elif isinstance(value, dict):
                self._merge(value, dst[key])
            elif key == "description":
                dst[key] = "{}\n{}".format(dst[key], value)
        return dst

    def _get_schema_object(self, base_dir, schema_path):

        json_path = "$..'%s'" % schema_path.split("/")[-1]
        schema_object = self._get_parser(json_path).find(self._content)
        if len(schema_object) == 0:
            schema_object = self._get_schema_object_from_file(
                base_dir, schema_path
            )
        else:
            schema_object = schema_object[0].value
        return schema_object

    def _get_schema_object_from_file(self, base_dir, schema_path):
        import yaml

        paths = schema_path.split("#")
        filename = os.path.join(base_dir, paths[0])
        filename = os.path.abspath(os.path.normpath(filename))
        with open(filename) as fid:
            schema_file = yaml.safe_load(fid)
        json_path = "$..'%s'" % schema_path.split("/")[-1]
        schema_object = self._get_parser(json_path).find(schema_file)[0].value
        return schema_object

    def _resolve_strings(self, content):
        """Fix up strings"""
        for key, value in content.items():
            if isinstance(value, dict):
                self._resolve_strings(value)
            elif key == "description":
                descr = copy.deepcopy(value)
                content[key] = Bundler.description(descr)

    def _resolve_license(self):
        """License object is not required by the OpenAPI spec.
        If the license object is provided then name is a required property.
        """
        if "license" not in self._content["info"]:
            self._content["info"]["license"] = {"name": "NO-LICENSE-PRESENT"}
        elif "name" not in self._content["info"]["license"]:
            raise Exception(
                "The following properties are REQUIRED: license.name"
            )

    def _resolve_keys(self, content):
        changes = {}
        for key, value in content.items():
            if isinstance(key, int):
                changes[key] = value
            if isinstance(value, dict):
                self._resolve_keys(value)
        for key, value in changes.items():
            content[str(key)] = value
            del content[key]


if __name__ == "__main__":
    Bundler(api_files=["./tests/api/api.yaml"]).bundle()



import subprocess
import os
from .openapiartplugin import OpenApiArtPlugin


class OpenApiArtProtobuf(OpenApiArtPlugin):
    def __init__(self, **kwargs):
        self._errors = []
        super(OpenApiArtProtobuf, self).__init__(**kwargs)
        self._filename = os.path.normpath(
            os.path.join(
                self._output_dir, "{}.proto".format(self._protobuf_file_name)
            )
        )
        self.default_indent = "  "
        self.proto_service_name = kwargs.get("proto_service", "Openapi")
        self.doc_dir = kwargs.get("doc_dir")
        self._init_fp(self._filename)

    def generate(self, openapi):
        self._errors = []
        self._openapi = openapi
        self._operations = {}
        self._write_header(self._openapi["info"])
        for name, schema_object in self._openapi["components"][
            "schemas"
        ].items():
            self._write_msg(name, schema_object)
        for name, response_object in self._openapi["components"][
            "responses"
        ].items():
            self._write_msg(name, response_object)
        for _, path_object in self._openapi["paths"].items():
            self._write_request_msg(path_object)
            self._write_response_msg(path_object)
        self._validate_error()
        self._write_service()
        self._close_fp()
        self.generate_doc()

    def generate_doc(self):
        if self.doc_dir is None:
            return
        process_args = [
            "protoc",
            "--doc_out={}".format(self.doc_dir),
            "--doc_opt=html,index.html",
            "--proto_path={}".format(self._output_dir),
            self._filename,
        ]
        cmd = " ".join(process_args)
        try:
            process = subprocess.Popen(cmd, shell=True)
            process.wait()
        except Exception:
            print("Bypassed generating proto document")
        # protoc --plugin=protoc-gen-doc=./protoc-gen-doc --doc_out=./doc --doc_opt=html,index.html sanity.proto

    def _validate_error(self):
        if len(self._errors) > 0:
            raise TypeError("\n".join(self._errors))

    def _get_operation(self, path_item_object):
        if "operationId" in path_item_object:
            operation_id = path_item_object["operationId"]
            if operation_id not in self._operations:
                operation = lambda x: None
                operation.rpc = self._get_camel_case(operation_id)
                operation.request = "google.protobuf.Empty"
                operation.response = "{}Response".format(operation.rpc)
                operation.stream = (
                    len(
                        self._get_parser('$.."application/octet-stream"').find(
                            path_item_object
                        )
                    )
                    > 0
                )
                self._operations[operation_id] = operation
            return self._operations[operation_id]
        return None

    def _write_request_msg(self, path_object):
        for _, path_item_object in path_object.items():
            operation = self._get_operation(path_item_object)
            if operation is None:
                continue
            if (
                len(self._get_parser("$..requestBody").find(path_item_object))
                > 0
            ):
                operation.request = "{}Request".format(operation.rpc)
                self._write()
                self._write("message {} {{".format(operation.request))
                for ref in self._get_parser('$..requestBody.."$ref"').find(
                    path_item_object
                ):
                    message = self._get_message_name(ref.value)
                    field_type = message.replace(".", "")
                    field_name = self._lowercase(field_type)
                    self._write(
                        "{} {} = 1;".format(field_type, field_name), indent=1
                    )
                self._write("}")

    def _write_response_msg(self, path_object):
        """
        application/octet-stream -> response (stream <operationId>Response)
        application/json -> response (<operationId>Response)
        """
        for _, path_item_object in path_object.items():
            operation = self._get_operation(path_item_object)
            if operation is None:
                continue
            for response in self._get_parser("$..responses").find(
                path_item_object
            ):
                response_fields = []
                for code, code_schema in response.value.items():
                    response_field = lambda: None
                    response_field.type = None
                    response_field.name = "status_code_{}".format(code)
                    response_field.field_uid = code_schema["x-field-uid"]
                    schema = self._get_parser("$..schema").find(
                        code_schema
                    )  # finds the first instance of schema in responses
                    if len(schema) > 0:
                        schema_ref = self._get_parser("$..'$ref'").find(
                            schema[0].value
                        )  # gets a ref
                    else:
                        schema_ref = self._get_parser("$..'$ref'").find(
                            code_schema
                        )  # gets a ref
                    if len(schema_ref) > 0:
                        schema = schema_ref[0].value
                    elif len(schema) > 0:
                        schema = schema[0].value
                    if "#/components/responses" in schema:
                        # lookup the response object and use the schema or ref in that object
                        jsonpath = "$.{}..schema".format(
                            schema[2:].replace("/", ".")
                        )
                        schema = (
                            self._get_parser(jsonpath)
                            .find(self._openapi)[0]
                            .value
                        )
                        ref = self._get_parser("$..'$ref'").find(schema)
                        if len(ref) > 0:
                            schema = ref[0].value
                    if "$ref" in schema:
                        response_field.type = self._get_message_name(
                            schema["$ref"]
                        ).replace(".", "")
                    elif "type" in schema:
                        response_field.type = self._get_message_name(
                            schema["type"]
                        ).replace(".", "")
                        if "format" in schema and schema["format"] == "binary":
                            response_field.type = "bytes"
                    elif len(schema) > 0:
                        response_field.type = self._get_message_name(
                            schema
                        ).replace(".", "")
                    else:
                        response_field.type = "string"
                    response_fields.append(response_field)
            self._write("message {} {{".format(operation.response))
            for response_field in response_fields:
                self._write(
                    "optional {} {} = {};".format(
                        response_field.type,
                        response_field.name,
                        response_field.field_uid,
                    ),
                    indent=1,
                )
            self._write("}")
            self._write()

    def _get_message_name(self, ref):
        return ref.split("/")[-1]

    def _write_header(self, info_object):
        self._write(
            self._justify_desc(
                self._info + "\n{}".format(self._license), use_multi=True
            )
        )
        self._write()
        self._write('syntax = "proto3";')
        self._write()
        self._write("package {};".format(self._protobuf_package_name))
        self._write()

        if self._go_sdk_package_dir is None:
            option_go_pkg = 'option go_package = "./{};{}";'.format(
                self._protobuf_package_name, self._protobuf_package_name
            )

        else:
            option_go_pkg = 'option go_package = "{}/{}";'.format(
                self._go_sdk_package_dir, self._protobuf_package_name
            )
        self._write(option_go_pkg)
        self._write()

        self._write('import "google/protobuf/descriptor.proto";')
        self._write('import "google/protobuf/empty.proto";')

    def _get_field_type(
        self,
        property_name,
        openapi_object,
        outer_format=None,
        outer_min=None,
        outer_max=None,
    ):
        """Convert openapi type -> protobuf type

        - type:number -> float
        - type:number [format: float] -> float
        - type:number [format: double] -> double
        - type:integer -> int32
        - type:integer [format:int32] -> int32
        - type:integer [format:int64] -> int64
        - type:integer [format:uint32] -> uint32
        - type:integer [format:uint64] -> uint64
        - type:boolean -> bool
        - type:string -> string
        - type:string [format:binary] -> bytes
        """
        if "type" in openapi_object:
            type = openapi_object["type"]
            if type == "boolean":
                return "bool"
            if type == "string":
                if "format" in openapi_object:
                    if (
                        str(outer_format) == "binary"
                        or openapi_object["format"] == "binary"
                    ):
                        return "bytes"
                elif "x-enum" in openapi_object:
                    enum_msg = self._camelcase("{}".format(property_name))
                    self._write_x_enum_msg(
                        enum_msg,
                        openapi_object["x-enum"],
                        property_name,
                        openapi_object,
                    )
                    return enum_msg + ".Enum"
                return "string"
            if type == "integer":
                type_format = openapi_object.get("format")
                min = openapi_object.get("minimum")
                max = openapi_object.get("maximum")
                if outer_min is not None:
                    min = outer_min
                if outer_max is not None:
                    max = outer_max
                if outer_format is not None:
                    type_format = outer_format
                return _get_integer_format(type_format, min, max)
            if type == "number":
                if "format" in openapi_object:
                    if (
                        str(outer_format) == "double"
                        or openapi_object["format"] == "double"
                    ):
                        return "double"
                    elif (
                        str(outer_format) == "float"
                        or openapi_object["format"] == "float"
                    ):
                        return "float"
                return "float"
            if type == "array":
                # TODO: outer min/max shouldn't exist for arrays
                type_format = openapi_object.get("format")
                min = openapi_object.get("minimum")
                max = openapi_object.get("maximum")

                item_type = self._get_field_type(
                    property_name,
                    openapi_object["items"],
                    outer_format=type_format,
                    outer_min=min,
                    outer_max=max,
                )

                if item_type == "integer":
                    item_type = _get_integer_format(type_format, min, max)
                return "repeated " + item_type
        elif "$ref" in openapi_object:
            return openapi_object["$ref"].split("/")[-1].replace(".", "")

    def _camelcase(self, value):
        camel_case = ""
        for piece in value.split("_"):
            camel_case += "{}{}".format(piece[0].upper(), piece[1:])
        return camel_case

    def _camelcase_to_snakecase(self, value, lower=False):
        word = ""
        insert_underscore = False

        for c in value:
            if c.isupper() or c.isdigit():
                if insert_underscore:
                    word += "_" + (c.lower() if lower else c)
                    insert_underscore = False
                else:
                    word += c.lower() if lower else c
            else:
                word += c if lower else c.upper()
                insert_underscore = True

        return word

    def _uppercase(self, value):
        return self._camelcase_to_snakecase(value, lower=False)

    def _lowercase(self, value):
        return self._camelcase_to_snakecase(value, lower=True)

    def _get_description(self, openapi_object):
        if "description" in openapi_object:
            return openapi_object["description"].replace('"', "")
        else:
            return "Description missing in models"

    def _write_x_enum_msg(
        self, enum_msg_name, enums, property_name, property_object
    ):
        """Follow google developers style guide for enums
        - reference: https://developers.google.com/protocol-buffers/docs/style#enums
        """
        self._write(
            "message {} {{".format(enum_msg_name.replace(".", "")), indent=1
        )
        self._write("enum Enum {", indent=2)
        if "unspecified" not in enums:
            self._write("{} = {};".format("unspecified", 0), indent=3)
        for key, value in enums.items():
            field_uid = value["x-field-uid"]
            self._write("{} = {};".format(key.lower(), field_uid), indent=3)
        self._write("}", indent=2)
        self._write("}", indent=1)

    def _write_msg(self, name, schema_object):
        msg_name = name.replace(".", "")
        print("writing msg {}".format(msg_name))
        self._write()
        self._write(self._justify_desc(self._get_description(schema_object)))
        self._write("message {} {{".format(msg_name), indent=0)
        if "content" in schema_object:
            # when accessing components/responses
            self._write_response_fields(msg_name, schema_object)
        else:
            # when accessing components/schemas
            self._write_msg_fields(name, schema_object)
        self._write("}")

    def _write_response_fields(self, msg_name, schema_object):
        try:
            ref = schema_object["content"]["application/json"]["schema"][
                "$ref"
            ]
            field_type = self._get_message_name(ref).replace(".", "")
            self._write(
                "{} {} = 1;".format(field_type, self._lowercase(field_type)),
                indent=1,
            )
        except AttributeError as err:
            print("Failed writing response {}: {}".format(msg_name, err))

    def _write_msg_fields(self, name, schema_object):
        if "properties" not in schema_object:
            return
        for property_name, property_object in schema_object[
            "properties"
        ].items():
            self._write()
            property_type = self._get_field_type(
                property_name, property_object
            )
            default = None
            if "default" in property_object:
                default = property_object["default"]
            if property_type.endswith(".Enum"):
                if default is not None:
                    default = "{}.{}".format(
                        property_type.split(" ")[-1], default.lower()
                    )
            if (
                "required" in schema_object
                and property_name in schema_object["required"]
                or property_type.startswith("repeated")
            ):
                optional = ""
            else:
                optional = "optional "
            desc = self._get_description(property_object)
            if default is not None:
                desc += "\ndefault = {}".format(default)
            if (
                optional == ""
                and property_type.startswith("repeated") is not True
            ):
                desc += "\nrequired = true"
            self._write(self._justify_desc(desc, indent=1))
            field_uid = property_object["x-field-uid"]
            self._write(
                "{}{} {} = {};".format(
                    optional, property_type, property_name.lower(), field_uid
                ),
                indent=1,
            )

    def _write_service(self):
        self._write()
        paths_object = self._openapi["paths"]
        self._write(
            self._justify_desc(self._get_description(paths_object), indent=1)
        )
        self._write("service {name} {{".format(name=self.proto_service_name))
        self._write()
        for url, path_object in self._openapi["paths"].items():
            for method, path_item_object in path_object.items():
                if method in ["get", "patch", "post", "delete"]:
                    self._write_rpc(url, method, path_item_object)
        self._write("}")

    def _write_rpc(self, url, method, path_item_object):
        """ """
        operation = self._get_operation(path_item_object)
        print("writing rpc {}".format(operation.rpc))
        self._write(
            self._justify_desc(
                self._get_description(path_item_object), indent=1
            )
        )
        line = "rpc {}({}) returns ({}{});".format(
            operation.rpc, operation.request, "", operation.response
        )
        self._write(line, indent=1)


def _min_max_in_range(format, min, max):
    if min is None or max is None:
        return True
    elif min is None or max is None:
        if min is None:
            min = max
        if max is None:
            max = min

    if min > max:
        return False

    if format == "uint32":
        return min >= 0 and max <= 4294967295
    elif format == "uint64":
        return min >= 0 and max <= 18446744073709551615
    elif format == "int32":
        return min >= -2147483648 and max <= 2147483647
    elif format == "int64":
        return min >= -9223372036854775808 and max <= 9223372036854775807
    else:
        raise Exception("unsupported integer format: {}".format(format))


def _format_from_range(min, max):
    if min is None and max is None:
        return "int32"
    elif min is None or max is None:
        if min is None:
            min = max
        if max is None:
            max = min

    if max < min:
        raise Exception("min %d cannot be less than max %d", min, max)

    # TODO: this logic needs to be replaced with the one commented out below
    if min > 2147483647 or max > 2147483647:
        return "int64"
    else:
        return "int32"
    # TODO: following snippet is more accurate but introduces breaking
    # changes and hence commented out
    # if self._min_max_in_range("uint32", min, max):
    #     return "uint32"
    # if self._min_max_in_range("uint64", min, max):
    #     return "uint64"
    # if self._min_max_in_range("int32", min, max):
    #     return "int32"
    # if self._min_max_in_range("int64", min, max):
    #     return "int64"


def _get_integer_format(type_format, min, max):
    valid_formats = ["int32", "int64", "uint32", "uint64"]
    if type_format is not None:
        if type_format in valid_formats:
            if not _min_max_in_range(type_format, min, max):
                raise Exception(
                    "format {} is not compatible with [min,max] [{},{}]".format(
                        type_format, min, max
                    )
                )
            return type_format
        raise Exception(
            "unsupported format %s, supported formats are: %s",
            type_format,
            valid_formats,
        )
    return _format_from_range(min, max)



def pascal_case(value: str) -> str:
    pascal_case = ""
    for piece in value.split("_"):
        pascal_case += "{}{}".format(piece[0].upper(), piece[1:])
    return pascal_case


def camel_case(value: str) -> str:
    value = pascal_case(value)
    return value[0].lower() + value[1:]



import os
import shutil
import openapiart.goserver.generator_context as ctx
from openapiart.goserver.go_interface_generator import (
    GoServerInterfaceGenerator,
)
from openapiart.goserver.go_controller_generator import (
    GoServerControllerGenerator,
)


class GoServerGenerator(object):
    def __init__(
        self,
        openapi,  # openapi.yaml.yaml
        output_root_path: str,
        module_path: str,
        models_prefix: str = "",
        models_path: str = "",
    ):
        self._output_root_path = output_root_path
        self._openapi = openapi
        self._context = ctx.GeneratorContext(openapi)
        self._context.output_path = os.path.join(output_root_path, "httpapi")
        if len(models_prefix) > 0:
            models_prefix = models_prefix + "."
        if len(models_path) == 0:
            models_path = module_path
        self._context.module_path = module_path
        self._context.models_prefix = models_prefix
        self._context.models_path = models_path
        print(
            "GoServer output directory: {}".format(self._context.output_path)
        )

    def generate(self):
        self._loadyaml()
        self._copy_static_files()
        GoServerInterfaceGenerator(self._context).generate()
        GoServerControllerGenerator(self._context).generate()

    def _loadyaml(self):
        # load servers
        if "servers" in self._openapi:
            servers = self._openapi["servers"]
            self._load_servers(servers)
        # load components
        if "components" in self._openapi:
            components = self._openapi["components"]
            if "schemas" in components:
                self._load_components(components["schemas"])
        # load routes
        if "paths" in self._openapi:
            for url, pathobj in self._openapi["paths"].items():
                self._loadroute(url, pathobj)

    def _load_servers(self, servers):
        for server in servers:
            s = ctx.Server(server)
            self._context.servers.append(s)

    def _load_components(self, components):
        for componentname, componentobj in components.items():
            c = ctx.Component(componentname, componentobj, self._context)
            self._context.components.append(c)

    def _loadroute(self, url: str, pathobj):
        http_methods = ["get", "post", "put", "delete", "head", "patch"]
        for methodname, methodobj in pathobj.items():
            if methodname not in http_methods:
                continue
            if "tags" not in methodobj:
                raise AttributeError(
                    "controller name missing from '{url} - {methodname}:'\nUse tags: [<name>]".format(
                        url=url, methodname=methodname
                    )
                )
            controllername = methodobj["tags"][0]
            ctrl = self._context.find_controller(controllername)
            ctrl.add_route(url, methodname, methodobj)

    def _copy_static_files(self):
        output_path = self._context.output_path
        if os.path.exists(output_path) is False:
            os.makedirs(output_path)
        srcfolder = os.path.dirname(__file__)
        name = "http_setup.go"
        shutil.copyfile(
            os.path.join(srcfolder, name), os.path.join(output_path, name)
        )
        print("copy: " + os.path.join(output_path, name))
        name = "response.go"
        shutil.copyfile(
            os.path.join(srcfolder, name), os.path.join(output_path, name)
        )
        print("copy: " + os.path.join(output_path, name))



import pytest


def test_req_choice_obj(api):
    config = api.prefix_config()
    req_obj = config.required_choice_object
    int_obj = req_obj.intermediate_obj
    int_obj.leaf.name = "some string"


def test_req_choice_validation_errors(api):
    conf = api.prefix_config()
    conf.required_object.e_a = 10
    conf.required_object.e_b = 20
    conf.a = "abc"
    conf.b = 10.2
    conf.c = 30

    conf.required_choice_object
    parent_exception = "choice is a mandatory property of <class 'sanity.sanity.RequiredChoiceParent'> and should not be set to None"

    with pytest.raises(Exception) as execinfo:
        conf.serialize(conf.DICT)
    assert execinfo.value.args[0] == parent_exception


def test_req_choice_serilize_deserialize(api):
    conf = api.prefix_config()
    conf.required_object.e_a = 10
    conf.required_object.e_b = 20
    conf.a = "abc"
    conf.b = 10.2
    conf.c = 30

    conf.required_choice_object.intermediate_obj.leaf.name = "some string"

    s_obj = conf.serialize(conf.DICT)

    conf.deserialize(s_obj)
    assert (
        conf.required_choice_object.intermediate_obj.leaf.name == "some string"
    )


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])



from google.protobuf import json_format


def test_protobuf(config, pb_config, utils):
    """
    1. Take a json without defaults and deserialize to snappi Config object.
    2. serialize the snappi config object so that defaults get added.
    3. Deserialize snappi json to protobuf.
    4. Serialize protobuf to json

    Validation :
        Compare jsons serialized from snappi & protobuf
    """
    # Take a json without defaults and deserialize to snappi Config object
    with open(utils.get_test_config_path("config.json")) as f:
        config.deserialize(f.read())

    # Serialize the snappi config object so that defaults get added
    snappi_json = config.serialize(config.JSON)

    # Deserialize snappi_json to protobuf
    pb_obj = json_format.Parse(snappi_json, pb_config)

    # Serialize protobuf to json
    pb_json = json_format.MessageToJson(
        pb_obj, preserving_proto_field_name=True
    )

    # Compare jsons serialized from snappi & protobuf
    assert utils.compare_json(snappi_json, pb_json)



import pytest


def test_remove(api):
    config = api.prefix_config()
    config.g.add(name="a")
    config.g.add(name="b")
    config.g.add(name="c")
    assert len(config.g) == 3
    assert config.g[0].name == "a"
    assert config.g[1].name == "b"
    assert config.g[2].name == "c"

    config.g.remove(1)
    config.g.add(name="d")
    assert len(config.g) == 3
    assert config.g[0].name == "a"
    assert config.g[1].name == "c"
    assert config.g[2].name == "d"


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])



import pytest


@pytest.mark.skip(reason="shall be restored")
def test_unique(config):

    # Update: There is no global and local diff
    # everything is considered as global now
    # *************** global unique ****************
    # Two similar objects with same Name.
    config.w_list.wobject(w_name="global_unique_similar_obj")
    config.w_list.wobject(w_name="global_unique_similar_obj")
    try:
        config.validate()
        pytest.fail("validation failed")
    except Exception as e:

        if "global_unique_similar_obj already exists" not in str(e):
            pytest.fail("global_unique_similar_obj validation failed")

    # Two similar objects with different name
    config.w_list[1].w_name = "global_unique_similar_obj1"
    config.validate()

    # Two different objects with same name
    config.name = "global_unique"
    config.w_list.wobject(w_name="global_unique")
    try:
        config.validate()
        pytest.fail("validation failed")
    except Exception as e:
        if "global_unique already exists" not in str(e):
            pytest.fail("global_unique validation failed")

    #  Two different objects with different name
    config.name = "global_unique1"
    config.validate()

    # *************** local unique ****************
    # Two similar objects with same Name.
    config.x_list.zobject(name="local_unique_similar")
    config.x_list.zobject(name="local_unique_similar")
    try:
        config.validate()
        pytest.fail("validation failed")
    except Exception as e:
        if "local_unique_similar already exists" not in str(e):
            pytest.fail("local_unique_similar validation failed")

    # Two similar objects with different name
    config.x_list[1].name = "local_unique_similar1"
    config.validate()

    # Two different objects with same name
    config.name = "local_global_mix"
    config.x_list.zobject(name="local_global_mix")
    try:
        config.validate()
        pytest.fail("validation failed")
    except Exception as e:
        if "local_global_mix already exists" not in str(e):
            pytest.fail("local_unique_similar validation failed")
    try:
        config.serialize()
        pytest.fail("serialization failed")
    except Exception as e:
        if "local_global_mix already exists" not in str(e):
            pytest.fail("local_unique_similar validation failed")
    # **********************************************

    # config.z_object.name = "local_unique"
    # config.y_object.y_name = "123"
    # config.x_list.zobject(name="local_unique")
    # try:
    #     config.validate()
    # except Exception as e:
    #     if "local_unique already exists" not in str(
    #         e
    #     ) or "y_name is not a valid type" not in str(e):
    #         pytest.fail("validation failed")
    # try:
    #     config.serialize()
    # except Exception as e:
    #     if "local_unique already exists" not in str(
    #         e
    #     ) or "y_name is not a valid type" not in str(e):
    #         pytest.fail("validation failed")

    # config.x_list[0].name = "local_unique_1"
    # config.y_object.y_name = "local_unique_1"
    # data = config.serialize(config.DICT)
    # data["y_object"]["y_name"] = "local_unique_decode"
    # data["z_object"]["name"] = "local_unique_decode"
    # try:
    #     config.deserialize(data)
    # except Exception as e:
    #     if "local_unique_decode already exists" not in str(
    #         e
    #     ) or "y_name is not a valid type" not in str(e):
    #         pytest.fail("validation failed")

    # config.name = "global_local_same_name_check"
    # config.x_list.zobject(name="global_local_same_name_check")
    # try:
    #     config.validate()
    # except Exception as e:
    #     pytest.fail("validation failed\n {e}".format(e=e))



import os
import subprocess


class GoTidy(object):
    def __init__(
        self,
        output_root_path,
    ):
        self._output_root_path = output_root_path

    def goTidy(self):
        print(
            "GoTidy output directory: {path}".format(
                path=self._output_root_path
            )
        )
        self._format_go()
        self._tidy_mod()

    def _format_go(self):
        """Format the generated go code"""
        try:
            process_args = [
                "goimports",
                "-w",
                self._output_root_path,
            ]
            cmd = " ".join(process_args)
            print("Formatting generated go files in folder: {}".format(cmd))
            process = subprocess.Popen(
                cmd, cwd=self._output_root_path, shell=True
            )
            process.wait()
        except Exception as e:
            print("Bypassed formatting of generated go ux file: {}".format(e))

    def _tidy_mod(self):
        """Tidy the mod file"""
        try:
            process_args = [
                "go",
                "mod",
                "tidy",
            ]
            os.environ["GO111MODULE"] = "on"
            print(
                "Tidying the generated go mod file: {}".format(
                    " ".join(process_args)
                )
            )
            process = subprocess.Popen(
                process_args,
                cwd=self._output_root_path,
                shell=False,
                env=os.environ,
            )
            process.wait()
        except Exception as e:
            print("Bypassed tidying the generated mod file: {}".format(e))



import pytest
import json


def test_grpc_set_config(utils, grpc_api):
    with open(utils.get_test_config_path("config.json")) as f:
        payload = json.load(f)
    result = grpc_api.set_config(payload)
    assert result.read() == b"success"


def test_grpc_get_config(utils, grpc_api):
    with open(utils.get_test_config_path("config.json")) as f:
        payload = json.load(f)
    result = grpc_api.get_config()
    assert result.a == payload.get("a")


def test_invalid_transport():
    with pytest.raises(Exception) as execinfo:
        pytest.module.api(location="localhost", transport="NetConf")
    assert (
        execinfo.value.args[0]
        == """NetConf is not within valid transport types ['http', 'grpc']"""
    )


def test_transport_with_ext():
    with pytest.raises(Exception) as execinfo:
        pytest.module.api(
            location="localhost",
            transport=pytest.module.Transport.GRPC,
            ext="ixnetwork",
        )
    assert (
        execinfo.value.args[0]
        == """ext and transport are not mutually exclusive. Please configure one of them."""
    )


def test_grpc_valid_version_check(utils, grpc_api):
    with open(utils.get_test_config_path("config.json")) as f:
        payload = json.load(f)
    try:
        grpc_api._version_check = True
        result = grpc_api.set_config(payload)
        assert result.read() == b"success"
    finally:
        grpc_api._version_check = False


def test_grpc_valid_inversion_check(utils, grpc_api):
    with open(utils.get_test_config_path("config.json")) as f:
        payload = json.load(f)
    try:
        grpc_api.get_local_version().api_spec_version = "0.2.1"
        grpc_api._version_check = True
        grpc_api.set_config(payload)
        raise Exception("expected version error")
    except Exception:
        pass
    finally:
        grpc_api.get_local_version().api_spec_version = "0.1.0"
        grpc_api._version_check = False


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])



import json
import os


CONFIGS_DIR = "tests/json_configs"


def compare_json(d1, d2):
    d1 = sorting(json.loads(d1))
    d2 = sorting(json.loads(d2))
    if d1 == d2:
        return True
    return False


def sorting(item):
    if isinstance(item, dict):
        return sorted((key, sorting(values)) for key, values in item.items())
    if isinstance(item, list):
        return sorted(sorting(x) for x in item)
    else:
        return item


def get_root_dir():
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_test_config_path(config_name):
    return os.path.join(
        os.path.dirname(get_root_dir()), CONFIGS_DIR, config_name
    )



from flask import Flask, request, Response
import threading
import json
import os
import importlib
import sys


app = Flask(__name__)
app.CONFIG = None
app.PACKAGE = None
app.UPDATE_CONFIG = None
app.PORT = 8444
app.HOST = "0.0.0.0"


@app.route("/api/config", methods=["POST"])
def set_config():
    config = app.PACKAGE.Api().prefix_config()
    config.deserialize(request.data.decode("utf-8"))
    test = config.h
    if test is not None and isinstance(test, bool) is False:
        return Response(
            status=590,
            response=json.dumps({"detail": "invalid data type"}),
            headers={"Content-Type": "application/json"},
        )
    else:
        app.CONFIG = config
        return Response(status=200)


@app.route("/api/config", methods=["PATCH"])
def update_configuration():
    config = app.PACKAGE.Api().update_config()
    config.deserialize(request.data.decode("utf-8"))
    app.UPDATE_CONFIG = config
    return Response(status=200)


@app.route("/api/config", methods=["GET"])
def get_config():
    app.CONFIG.a = "asdf"
    app.CONFIG.b = 1.1
    app.CONFIG.c = 1
    app.CONFIG.required_object.e_a = 1.1
    app.CONFIG.required_object.e_b = 1.2
    serialized_config = app.CONFIG.serialize()
    return Response(serialized_config, mimetype="application/json", status=200)


@app.route("/api/capabilities/version", methods=["GET"])
def get_version():
    return Response(
        app.PACKAGE.Api().get_local_version().serialize(),
        mimetype="application/json",
        status=200,
    )


@app.after_request
def after_request(resp):
    print(request.method, request.url, " -> ", resp.status)
    return resp


class OpenApiServer(object):
    def __init__(self, package):
        # TODO Shall change the below sanity path to be dynamic
        pkg_name = "sanity"
        lib_path = "../../art/%s" % pkg_name
        sys.path.append(
            os.path.join(os.path.join(os.path.dirname(__file__), lib_path))
        )
        app.PACKAGE = importlib.import_module(pkg_name)
        app.CONFIG = app.PACKAGE.Api().prefix_config()

    @staticmethod
    def web_server():
        app.run(host=app.HOST, port=app.PORT, debug=False, use_reloader=False)

    def start(self):
        self._web_server_thread = threading.Thread(
            target=OpenApiServer.web_server
        )
        self._web_server_thread.setDaemon(True)
        self._web_server_thread.start()
        return self



import os
import re
from jsonpath_ng import parse
import openapiart.goserver.string_util as util
import openapiart.goserver.generator_context as ctx
from openapiart.goserver.writer import Writer


class GoServerControllerGenerator(object):
    def __init__(self, context):
        # type: (ctx.GeneratorContext) -> None
        self._indent = "\t"
        self._root_package = context.module_path
        self._package_name = "controllers"
        self._ctx = context
        self._output_path = os.path.join(context.output_path, "controllers")

    def generate(self):
        self._write_controllers()

    def _write_controllers(self):
        if not os.path.exists(self._output_path):
            os.makedirs(self._output_path)
        for ctrl in self._ctx.controllers:
            self._write_controller(ctrl)

    def _write_controller(self, ctrl):
        # type: (ctx.Controller) -> None
        filename = ctrl.yamlname.lower() + "_controller.go"
        fullname = os.path.join(self._output_path, filename)
        w = Writer(self._indent)
        self._write_header(w)
        self._write_import(w)
        self._write_controller_struct(w, ctrl)
        self._write_newcontroller(w, ctrl)
        self._write_routes(w, ctrl)
        self._write_methods(w, ctrl)
        with open(fullname, "w") as file:
            print("Controller: {}".format(fullname))
            for line in w.strings:
                file.write(line + "\n")

    def _write_header(self, w):
        # type: (Writer) -> None
        w.write_line(
            """// This file is autogenerated. Do not modify
            package {package_name}
            """.format(
                package_name=self._package_name
            )
        )

    def _write_import(self, w):
        # type: (Writer) -> None
        w.write_line("import (").push_indent().write_line(
            """\"io"
            "net/http"
            "google.golang.org/protobuf/encoding/protojson"
            "{root_package}/httpapi"
            "{root_package}/httpapi/interfaces"
            {models_prefix} "{models_path}\"""".format(
                root_package=self._root_package,
                models_prefix=re.sub("[.]", "", self._ctx.models_prefix),
                models_path=self._ctx.models_path,
            )
        ).pop_indent().write_line(")", "")

    def _struct_name(self, ctrl):
        # type: (ctx.Controller) -> str
        return util.camel_case(ctrl.controller_name)

    def _write_controller_struct(self, w, ctrl):
        # type: (Writer, ctx.Controller) -> None
        w.write_line(
            "type {name} struct {{".format(name=self._struct_name(ctrl))
        )
        w.push_indent()
        w.write_line(
            "handler interfaces.{name}".format(name=ctrl.service_handler_name)
        )
        w.pop_indent()
        w.write_line("}", "")

    def _write_newcontroller(self, w, ctrl):
        # type: (Writer, ctx.Controller) -> None
        w.write_line(
            "func NewHttp{ctl_name}(handler interfaces.{handle_name}) interfaces.{ctl_name} {{".format(
                ctl_name=ctrl.controller_name,
                handle_name=ctrl.service_handler_name,
            )
        ).push_indent()
        w.write_line(
            "return &{name}{{handler}}".format(name=self._struct_name(ctrl))
        ).pop_indent()
        w.write_line("}", "")

    def _write_routes(self, w, ctrl):
        # type: (Writer, ctx.Controller) -> None
        w.write_line(
            "func (ctrl *{name}) Routes() []httpapi.Route {{".format(
                name=self._struct_name(ctrl)
            )
        ).push_indent()
        w.write_line("return [] httpapi.Route {").push_indent()
        for r in ctrl.routes:
            w.write_line(
                """{{ Path: "{url}", Method: "{method}", Name: "{operation_name}", Handler: ctrl.{operation_name}}},""".format(
                    url=r.url, method=r.method, operation_name=r.operation_name
                )
            )
        w.pop_indent()
        w.write_line(
            "}",
        ).pop_indent()
        w.write_line("}", "")

    def _write_methods(self, w, ctrl):
        # type: (Writer, ctx.Controller) -> None
        self._has_warning_check = False
        for route in ctrl.routes:
            self._write_method(w, ctrl, route)
        if self._has_warning_check:
            w.write_line(
                """var {name}MrlOpts = protojson.MarshalOptions{{
                    UseProtoNames:   true,
                    AllowPartial:    true,
                    EmitUnpopulated: true,
                    Indent:          "  ",
                }}""".format(
                    name=util.camel_case(ctrl.yamlname)
                )
            )

    def _write_method(self, w, ctrl, route):
        # type: (Writer, ctx.Controller,ctx.ControllerRoute) -> None
        w.write_line("/*")
        w.write_line(
            "{operation_name}: {method} {url}".format(
                operation_name=route.operation_name,
                method=route.method,
                url=route.url,
            )
        )
        w.write_line("Description: " + route.description)
        w.write_line("*/")
        w.write_line(
            "func (ctrl *{name}) {opt_name}(w http.ResponseWriter, r *http.Request) {{".format(
                name=self._struct_name(ctrl), opt_name=route.operation_name
            )
        )
        w.push_indent()
        request_body = route.requestBody()  # type: ctx.Component
        rsp_400_error = "response{}400".format(route.operation_name)
        rsp_500_error = "response{}500".format(route.operation_name)
        rsp_errors = [500]
        if request_body is not None:
            rsp_errors.append(400)
            modelname = request_body.model_name
            full_modelname = request_body.full_model_name
            new_modelname = self._ctx.models_prefix + "New" + modelname

            w.write_line(
                """var item {full_modelname}
                if r.Body != nil {{
                    body, readError := io.ReadAll(r.Body)
                    if body != nil {{
                        item = {new_modelname}()
                        err := item.FromJson(string(body))
                        if err != nil {{
                            ctrl.{rsp_400_error}(w, err)
                            return
                        }}
                    }} else {{
                        ctrl.{rsp_400_error}(w, readError)
                        return
                    }}
                }} else {{
                    bodyError := errors.New(\"Request do not have any body\")
                    ctrl.{rsp_500_error}(w, bodyError)
                    return
                }}
                result := ctrl.handler.{operation_name}(item, r)""".format(
                    full_modelname=full_modelname,
                    new_modelname=new_modelname,
                    rsp_400_error=rsp_400_error,
                    rsp_500_error=rsp_500_error,
                    operation_name=route.operation_name,
                )
            )
        else:
            w.write_line(
                "result := ctrl.handler.{name}(r)".format(
                    name=route.operation_name
                )
            )

        error_responses = []
        for response in route.responses:
            if int(response.response_value) in rsp_errors:
                error_responses.append(response)

            # no response content defined, return as 'any'
            if response.has_json:
                write_method = "WriteJSONResponse"
            elif response.has_binary:
                write_method = "WriteByteResponse"
            else:
                write_method = "WriteAnyResponse"

            # This is require as workaround of https://github.com/open-traffic-generator/openapiart/issues/220
            if self._need_warning_check(route, response):
                rsp_section = """data, err := {mrl_name}MrlOpts.Marshal(result.StatusCode200().Msg())
                        if err != nil {{
                            ctrl.{rsp_400_error}(w, err)
                        }}
                        httpapi.WriteCustomJSONResponse(w, 200, data)
                    """.format(
                    mrl_name=util.camel_case(ctrl.yamlname),
                    rsp_400_error=rsp_400_error,
                )
            else:
                rsp_section = """if _, err := httpapi.{write_method}(w, {response_value}, result.StatusCode{response_value}()); err != nil {{
                            log.Print(err.Error())
                    }}""".format(
                    write_method=write_method,
                    response_value=response.response_value,
                )

            w.write_line(
                """if result.HasStatusCode{response_value}() {{
                               {rsp_section}
                               return
                           }}""".format(
                    response_value=response.response_value,
                    rsp_section=rsp_section,
                )
            )
        w.write_line(
            'ctrl.{rsp_500_error}(w, errors.New("Unknown error"))'.format(
                rsp_500_error=rsp_500_error
            )
        )
        w.pop_indent()
        w.write_line("}", "")

        for err_rsp in error_responses:
            schema = parse("$..schema").find(err_rsp.response_obj)[0].value
            if "$ref" in schema:
                schema = self._ctx.get_object_from_ref(schema["$ref"])
            for prop_name, prop_value in schema["properties"].items():
                break
            prop_type = prop_value["type"]
            if prop_type not in ["string", "array"]:
                raise Exception(
                    "Expecting Errors are string/ array of strings"
                )
            if prop_type == "string":
                set_errors = """Set{prop_name}(rsp_err.Error())""".format(
                    prop_name=util.pascal_case(prop_name)
                )
            else:
                set_errors = (
                    """Set{prop_name}([]string{{rsp_err.Error()}})""".format(
                        prop_name=util.pascal_case(prop_name)
                    )
                )
            w.write_line(
                """func (ctrl *{struct_name}) {method_name}(w http.ResponseWriter, rsp_err error) {{
                result := {models_prefix}New{response_model_name}()
                result.StatusCode{response_value}().{set_errors}
                if _, err := httpapi.WriteJSONResponse(w, {response_value}, result.StatusCode{response_value}()); err != nil {{
                    log.Print(err.Error())
                }}
            }}
            """.format(
                    struct_name=self._struct_name(ctrl),
                    method_name=rsp_400_error
                    if int(err_rsp.response_value) == 400
                    else rsp_500_error,
                    models_prefix=self._ctx.models_prefix,
                    response_model_name=route.response_model_name,
                    response_value=err_rsp.response_value,
                    set_errors=set_errors,
                )
            )

    def _need_warning_check(self, route, response):
        parse_schema = parse("$..schema").find(response.response_obj)
        schema = [s.value for s in parse_schema]
        if len(schema) == 0:
            return False
        schema = schema[0]
        if "$ref" in schema:
            schema = self._ctx.get_object_from_ref(schema["$ref"])
        parse_warnings = parse("$..warnings").find(schema)
        if (
            route.method in ["PUT", "POST"]
            and int(response.response_value) == 200
            and response.has_json
            and len(parse_warnings) > 0
        ):
            self._has_warning_check = True
            return True
        return False

    # def _write_servicehandler_interface(self, w: Writer, ctrl: ctx.Controller):
    #     w.write_line(
    #         f"type {ctrl.service_handler_name} interface {{",
    #     )
    #     w.push_indent()
    #     w.write_line(
    #         f"GetController() {ctrl.controller_name}",
    #     )
    #     for r in ctrl.routes:
    #         response_model_name = r.operation_name + 'Response'
    #         w.write_line(
    #             f"{r.operation_name}(r *http.Request) models.{response_model_name}",
    #         )
    #     w.pop_indent()
    #     w.write_line(
    #         "}",
    #         ""
    #     )
    #     pass



import os
import re
import openapiart.goserver.string_util as util
import openapiart.goserver.generator_context as ctx
from openapiart.goserver.writer import Writer


class GoServerInterfaceGenerator(object):
    def __init__(self, context: ctx.GeneratorContext):
        self._indent = "\t"
        self._models_prefix = context.models_prefix
        self._root_package = context.module_path
        self._package_name = "interfaces"
        self._ctx = context
        self._output_path = os.path.join(context.output_path, "interfaces")

    def generate(self):
        self._write_interfaces()

    def _write_interfaces(self):
        if not os.path.exists(self._output_path):
            os.makedirs(self._output_path)
        for ctrl in self._ctx.controllers:
            self._write_interface(ctrl)

    def _write_interface(self, ctrl: ctx.Controller):
        filename = ctrl.yamlname.lower() + "_interface.go"
        fullname = os.path.join(self._output_path, filename)
        w = Writer(self._indent)
        self._write_header(w)
        self._write_import(w)
        self._write_path_param_const(w, ctrl)
        self._write_controller_interface(w, ctrl)
        self._write_servicehandler_interface(w, ctrl)
        with open(fullname, "w") as file:
            print(f"Interface: {fullname}")
            for line in w.strings:
                file.write(line + "\n")
            pass
        pass

    def _write_header(self, w: Writer):
        w.write_line(
            "// This file is autogenerated. Do not modify",
            f"package {self._package_name}",
            "",
        )

    def _write_import(self, w: Writer):
        w.write_line("import (").push_indent().write_line(
            '"net/http"',
            f'"{self._root_package}/httpapi"',
            f'{re.sub("[.]", "", self._ctx.models_prefix)} "{self._ctx.models_path}"',
        ).pop_indent().write_line(")", "")

    def _write_path_param_const(self, w: Writer, ctrl: ctx.Controller):
        params: [str] = []
        for r in ctrl.routes:
            for param in r.route_parameters:
                if param not in params:
                    params.append(param)
        if len(params) > 0:
            w.write_line("const (")
            w.push_indent()
            for param in params:
                w.write_line(
                    f'{util.pascal_case(ctrl.yamlname)}{util.pascal_case(param)} = "{param}"'
                )
            w.pop_indent()
            w.write_line(")", "")
        pass

    def _write_route_description(self, w: Writer, r: ctx.ControllerRoute):
        w.write_line("/*")
        w.write_line(f"{r.operation_name}: {r.method} {r.url}")
        w.write_line("Description: " + r.description)
        w.write_line("*/")

    def _write_controller_interface(self, w: Writer, ctrl: ctx.Controller):
        w.write_line(
            f"type {ctrl.controller_name} interface {{",
        )
        w.push_indent()
        w.write_line(
            "Routes() []httpapi.Route",
        )
        for r in ctrl.routes:
            # self._write_route_description(w, r)
            w.write_line(
                f"{r.operation_name}(w http.ResponseWriter, r *http.Request)",
            )
        w.pop_indent()
        w.write_line("}", "")

    def _write_servicehandler_interface(self, w: Writer, ctrl: ctx.Controller):
        w.write_line(
            f"type {ctrl.service_handler_name} interface {{",
        )
        w.push_indent()
        w.write_line(
            f"GetController() {ctrl.controller_name}",
        )
        for r in ctrl.routes:
            self._write_route_description(w, r)
            full_responsename = r.full_responsename
            request_body: Component = r.requestBody()
            if request_body is not None:
                full_requestname = request_body.full_model_name
                w.write_line(
                    f"{r.operation_name}(rbody {full_requestname}, r *http.Request) {full_responsename}"
                )
            else:
                w.write_line(
                    f"{r.operation_name}(r *http.Request) {full_responsename}"
                )
        w.pop_indent()
        w.write_line("}", "")



"""Abstract plugin class
"""
import jsonpath_ng


class OpenApiArtPlugin(object):
    """Abstract class for creating a plugin generator"""

    def __init__(self, **kwargs):
        self._fp = None
        self._license = kwargs["license"]
        self._info = kwargs["info"]
        self._output_dir = kwargs["output_dir"]
        self._python_module_name = (
            None
            if "python_module_name" not in kwargs
            else kwargs["python_module_name"]
        )
        self._protobuf_package_name = kwargs["protobuf_package_name"]
        self._protobuf_file_name = kwargs["protobuf_package_name"]
        self._go_sdk_package_dir = kwargs["go_sdk_package_dir"]
        self._generate_version_api = kwargs.get("generate_version_api")
        if self._generate_version_api is None:
            self._generate_version_api = False
        self._api_version = kwargs.get("api_version")
        if self._api_version is None:
            self._api_version = ""
        self._sdk_version = kwargs.get("sdk_version")
        if self._sdk_version is None:
            self._sdk_version = ""
        self._go_sdk_package_name = (
            None
            if "go_sdk_package_name" not in kwargs
            else kwargs["go_sdk_package_name"]
        )
        self.default_indent = "    "
        self._parsers = {}

    def _init_fp(self, filename):
        self._filename = filename
        self._fp = open(self._filename, "wb")

    def _close_fp(self):
        self._fp.close()

    def _write(self, line="", indent=0, newline=True):
        line = "{}{}{}".format(
            self.default_indent * indent, line, "\n" if newline else ""
        )
        self._fp.write(line.encode())

    def _get_parser(self, pattern):
        if pattern not in self._parsers:
            parser = jsonpath_ng.parse(pattern)
            self._parsers[pattern] = parser
        else:
            parser = self._parsers[pattern]
        return parser

    def _get_camel_case(self, value):
        camel_case = ""
        for piece in value.split("_"):
            camel_case += piece[0].upper()
            if len(piece) > 1:
                camel_case += piece[1:]
        return camel_case

    def _justify_desc(self, text, indent=0, use_multi=False):
        indent = " " * (indent * 2)
        lines = []
        text = text.split("\n")
        for line in text:
            char_80 = ""
            for word in line.split(" "):
                if len(char_80) <= 80:
                    char_80 += word + " "
                    continue
                lines.append(char_80.strip())
                char_80 = word + " "
            if char_80 != "":
                lines.append(char_80.strip())
            # lines.append("\n{}{}".format(indent, comment).join(each_line))
        if use_multi is True:
            return (
                "{}/* ".format(indent)
                + "\n{} * ".format(indent).join(lines)
                + " */"
            )
        return "{}// ".format(indent) + "\n{}// ".format(indent).join(lines)



"""Generator

Given an openapi.yaml file that has been produced by the Bundler class in the
bundler.py file the Generator class will produce an enhanced python ux file.

TBD:
- packet slicing using constants
- docstrings
- type checking
"""
from email import message
from logging import warning
import sys
import yaml
import os
import re
import requests
import pkgutil
import importlib
from jsonpath_ng import parse
from .openapiartplugin import OpenApiArtPlugin

MODELS_RELEASE = "v0.3.3"


class FluentRpc(object):
    """SetConfig(config Config) error"""

    def __init__(self):
        self.method = None
        self.operation_name = None
        self.request_class = None
        self.request_property = None
        self.good_response_type = None  # considering 200-ok
        self.bad_responses = []  # != 200-ok
        self.description = None
        self.http_method = None
        self.good_response_property = None
        self.x_status = None


class Generator:
    """Generates python classes based on an openapi.yaml file produced by the
    bundler.py infrastructure.
    """

    def __init__(
        self,
        openapi_filename,
        package_name,
        protobuf_package_name,
        output_dir=None,
        extension_prefix=None,
        generate_version_api=None,
        api_version=None,
        sdk_version=None,
    ):
        self._parsers = {}
        self._base_url = ""
        self._generated_methods = []
        self._generated_classes = []
        self._generated_top_level_factories = []
        self._openapi_filename = openapi_filename
        self._extension_prefix = extension_prefix
        self.__python = os.path.normpath(sys.executable)
        self.__python_dir = os.path.dirname(self.__python)
        self._src_dir = output_dir
        self._output_dir = os.path.join(output_dir, package_name)
        if os.path.exists(self._output_dir) is False:
            os.mkdir(self._output_dir)
        self._package_name = package_name
        self._protobuf_package_name = protobuf_package_name
        self._output_file = package_name
        self._docs_dir = os.path.join(self._src_dir, "..", "docs")
        self._deprecated_properties = {}
        self._generate_version_api = generate_version_api
        if self._generate_version_api is None:
            self._generate_version_api = False
        self._api_version = api_version
        if self._api_version is None:
            self._api_version = ""
        self._sdk_version = sdk_version
        if self._sdk_version is None:
            self._sdk_version = ""
        self._get_openapi_file()
        # self._plugins = self._load_plugins()

    def _get_parser(self, pattern):
        if pattern not in self._parsers:
            parser = parse(pattern)
            self._parsers[pattern] = parser
        else:
            parser = self._parsers[pattern]
        return parser

    def _load_plugins(self):
        plugins = []
        pkg_dir = os.path.dirname(__file__)
        for (_, name, _) in pkgutil.iter_modules([pkg_dir]):
            module_name = "openapiart." + name
            importlib.import_module(module_name)
            obj = sys.modules[module_name]
            for dir_name in dir(obj):
                if dir_name.startswith("_"):
                    continue
                dir_obj = getattr(obj, dir_name)
                print(dir_obj)
                if issubclass(dir_obj.__class__, OpenApiArtPlugin):
                    plugins.append(dir_obj)
        return plugins

    def _get_openapi_file(self):
        if self._openapi_filename is None:
            OPENAPI_URL = (
                "https://github.com/open-traffic-generator/models/releases"
                "/download/%s/openapi.yaml"
            ) % MODELS_RELEASE
            response = requests.request(
                "GET", OPENAPI_URL, allow_redirects=True
            )
            if response.status_code != 200:
                raise Exception(
                    "Unable to retrieve the Open Traffic Generator openapi.yaml"
                    " file [%s]" % response.content
                )
            openapi_content = response.content

            project_dir = os.path.dirname(os.path.dirname(__file__))
            with open(os.path.join(project_dir, "models-release"), "w") as out:
                out.write(MODELS_RELEASE)
        else:
            with open(self._openapi_filename, "rb") as fp:
                openapi_content = fp.read()
        self._openapi = yaml.safe_load(openapi_content)
        self._openapi_version = self._openapi["info"]["version"]
        print("generating using model version %s" % self._openapi_version)

    def generate(self):
        self._base_url = ""
        self._get_base_url()
        self._api_filename = os.path.join(
            self._output_dir, self._output_file + ".py"
        )
        with open(self._api_filename, "w") as self._fid:
            self._fid.write(
                "# {} {}\n".format(
                    self._openapi["info"]["title"],
                    self._openapi_version,
                )
            )
            self._fid.write(
                "# License: {}\n".format(
                    self._openapi["info"]["license"]["name"]
                ),
            )
            self._fid.write("\n")
        with open(
            os.path.join(os.path.dirname(__file__), "common.py"), "r"
        ) as fp:
            common_content = fp.read()
            cnf_text = "import sanity_pb2_grpc as pb2_grpc"
            modify_text = "try:\n    from {pkg_name} {text}\nexcept ImportError:\n    {text}".format(
                pkg_name=self._package_name,
                text=cnf_text.replace("sanity", self._protobuf_package_name),
            )
            common_content = common_content.replace(cnf_text, modify_text)

            cnf_text = "import sanity_pb2 as pb2"
            modify_text = "try:\n    from {pkg_name} {text}\nexcept ImportError:\n    {text}".format(
                pkg_name=self._package_name,
                text=cnf_text.replace("sanity", self._protobuf_package_name),
            )
            common_content = common_content.replace(cnf_text, modify_text)

            if re.search(r"def[\s+]api\(", common_content) is not None:
                self._generated_top_level_factories.append("api")
            if self._extension_prefix is not None:
                common_content = common_content.replace(
                    r'"{}_{}".format(__name__, ext)',
                    r'"'
                    + self._extension_prefix
                    + r"_{}."
                    + self._package_name
                    + r'_api".format(ext)',
                )
        with open(self._api_filename, "a") as self._fid:
            self._fid.write(common_content)
        methods, factories, rpc_methods = self._get_methods_and_factories()
        self._write_api_class(methods, factories)
        self._write_http_api_class(methods)
        self._write_rpc_api_class(rpc_methods)
        self._write_init()
        return self

    def _get_base_url(self):
        self._base_url = ""
        if "servers" in self._openapi:
            server = self._openapi["servers"][0]
            try:
                self._base_url = server["variables"]["basePath"]["default"]
                if not self._base_url.startswith("/"):
                    self._base_url = "/" + self._base_url
            except KeyError:
                pass

    def _write_init(self):
        filename = os.path.join(self._output_dir, "__init__.py")
        with open(filename, "w") as self._fid:
            for class_name in self._generated_classes:
                self._write(
                    0, "from .%s import %s" % (self._output_file, class_name)
                )
            for factory_name in self._generated_top_level_factories:
                self._write(
                    0, "from .%s import %s" % (self._output_file, factory_name)
                )

    def _find(self, path, schema_object):
        finds = self._get_parser(path).find(schema_object)
        for find in finds:
            yield find.value
            self._get_parser(path).find(find.value)

    def _get_methods_and_factories(self):
        """
        Parse methods and top level objects from yaml file to be later used in
        code generation.
        """
        methods = []
        rpc_methods = []
        factories = []
        refs = []
        self._top_level_schema_refs = []

        # parse methods
        for path in self._get_api_paths():
            operation = path["operation"]
            method_name = operation["operationId"].replace(".", "_").lower()
            if method_name in self._generated_methods:
                continue
            self._generated_methods.append(method_name)
            print("found method %s" % method_name)
            rpc = FluentRpc()
            rpc.method = method_name
            rpc.operation_name = self._get_external_struct_name(
                operation["operationId"].replace(".", "_")
            )
            rpc.description = self._get_description(operation)
            request = self._get_parser("$..requestBody..schema").find(
                operation
            )
            for req in request:
                (
                    _,
                    property_name,
                    class_name,
                    ref,
                ) = self._get_object_property_class_names(req.value)
                if ref:
                    refs.append(ref)
                    rpc.request_property = property_name
                    rpc.request_class = class_name

            response_type = None
            response_list = operation.get("responses")
            if response_list is None:
                raise Exception("{} should have responses".format(method_name))
            for response_code, response_property in response_list.items():
                if int(response_code) == 200:
                    rpc.good_response_property = response_property
                    schema_obj = self._get_parser("$..schema").find(
                        response_property
                    )
                    if len(schema_obj) == 0:
                        (
                            response_name,
                            _,
                            _,
                            ref,
                        ) = self._get_object_property_class_names(
                            response_property
                        )
                        if response_name is not None:
                            response = self._get_parser('$.."$ref"').find(
                                self._get_object_from_ref(ref)
                            )
                            if len(response) > 0:
                                (
                                    _,
                                    response_type,
                                    _,
                                    ref,
                                ) = self._get_object_property_class_names(
                                    response[0].value
                                )
                                if ref:
                                    refs.append(ref)
                    else:
                        (
                            _,
                            response_type,
                            _,
                            ref,
                        ) = self._get_object_property_class_names(
                            schema_obj[0].value
                        )
                        if ref:
                            refs.append(ref)
                else:
                    rpc.bad_responses.append(str(response_code))

            if response_type is None:
                # TODO: response type is usually None for schema which does not
                # contain any ref (e.g. response of POST /results/capture)
                pass

            rpc.good_response_type = response_type
            rpc.http_method = path["method"]
            # TODO: restore behavior
            if "x-status" in path["operation"] and path["operation"][
                "x-status"
            ].get("status") in ["deprecated", "under_review"]:
                rpc.x_status = (
                    path["operation"]["x-status"]["status"],
                    path["operation"]["x-status"]["information"],
                )
            methods.append(
                {
                    "name": method_name,
                    "args": ["self"]
                    if len(request) == 0
                    else ["self", "payload"],
                    "http_method": path["method"],
                    "url": self._base_url + path["url"],
                    "description": self._get_description(operation),
                    "response_type": response_type,
                    # TODO: restore behavior
                    "x_status": (
                        path["operation"]["x-status"]["status"],
                        path["operation"]["x-status"]["information"],
                    )
                    if path["operation"].get("x-status", {}).get("status")
                    in ["deprecated", "under_review"]
                    else None,
                }
            )
            rpc_methods.append(rpc)

        # parse top level objects (arguments for API requests)
        for ref in refs:
            if ref in self._generated_methods:
                continue
            self._generated_methods.append(ref)
            ret = self._get_object_property_class_names(ref)
            _, property_name, class_name, _ = ret
            schema_object = self._get_object_from_ref(ref)
            if "type" not in schema_object:
                continue
            print("found top level factory method %s" % property_name)
            if schema_object["type"] == "array":
                ref = schema_object["items"]["$ref"]
                _, _, class_name, _ = self._get_object_property_class_names(
                    ref
                )
                class_name = "%sIter" % class_name
                self._top_level_schema_refs.append((ref, property_name))
            self._top_level_schema_refs.append((ref, None))

            factories.append({"name": property_name, "class_name": class_name})

        for ref, property_name in self._top_level_schema_refs:
            if property_name is None:
                self._write_openapi_object(ref)
            else:
                self._write_openapi_list(ref, property_name)

        return methods, factories, rpc_methods

    def _write_rpc_api_class(self, rpc_methods):
        class_code = """class GrpcApi(Api):
    # OpenAPI gRPC Api
    def __init__(self, **kwargs):
        super(GrpcApi, self).__init__(**kwargs)
        self._stub = None
        self._channel = None
        self._request_timeout = 10
        self._keep_alive_timeout = 10 * 1000
        self._location = (
            kwargs["location"]
            if "location" in kwargs and kwargs["location"] is not None
            else "localhost:50051"
        )
        self._transport = kwargs["transport"] if "transport" in kwargs else None
        self._logger = kwargs["logger"] if "logger" in kwargs else None
        self._loglevel = kwargs["loglevel"] if "loglevel" in kwargs else logging.DEBUG
        if self._logger is None:
            stdout_handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter(fmt="%(asctime)s [%(name)s] [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
            formatter.converter = time.gmtime
            stdout_handler.setFormatter(formatter)
            self._logger = logging.Logger(self.__module__, level=self._loglevel)
            self._logger.addHandler(stdout_handler)
        self._logger.debug("gRPCTransport args: {}".format(", ".join(["{}={!r}".format(k, v) for k, v in kwargs.items()])))

    def _get_stub(self):
        if self._stub is None:
            CHANNEL_OPTIONS = [('grpc.enable_retries', 0),
                               ('grpc.keepalive_timeout_ms', self._keep_alive_timeout)]
            self._channel = grpc.insecure_channel(self._location, options=CHANNEL_OPTIONS)
            self._stub = pb2_grpc.OpenapiStub(self._channel)
        return self._stub

    def _serialize_payload(self, payload):
        if not isinstance(payload, (str, dict, OpenApiBase)):
            raise Exception("We are supporting [str, dict, OpenApiBase] object")
        if isinstance(payload, OpenApiBase):
            payload = payload.serialize()
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        return payload

    @property
    def request_timeout(self):
        \"\"\"duration of time in seconds to allow for the RPC.\"\"\"
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, timeout):
        self._request_timeout = timeout

    @property
    def keep_alive_timeout(self):
        return self._keep_alive_timeout

    @keep_alive_timeout.setter
    def keep_alive_timeout(self, timeout):
        self._keep_alive_timeout = timeout * 1000

    def close(self):
        if self._channel is not None:
            self._channel.close()
            self._channel = None
            self._stub = None"""

        self._generated_classes.append("Transport")
        with open(self._api_filename, "a") as self._fid:
            self._write()
            self._write()
            self._write(0, class_code)
            for rpc_method in rpc_methods:
                self._write()
                status_msg = ""
                if rpc_method.x_status is not None:
                    status_msg = "%s api is %s, %s" % (
                        rpc_method.method,
                        rpc_method.x_status[0],
                        rpc_method.x_status[1],
                    )
                    key = "{}.{}".format("GrpcApi", rpc_method.method)
                    self._deprecated_properties[key] = rpc_method.x_status[1]

                if rpc_method.request_class is None:
                    self._write(1, "def %s(self):" % rpc_method.method)
                    if status_msg != "":
                        self._write(2, "self.add_warnings('%s')" % status_msg)
                    self._write(2, "stub = self._get_stub()")
                    self._write(
                        2,
                        "empty = pb2_grpc.google_dot_protobuf_dot_empty__pb2.Empty()",
                    )
                    self._write(
                        2,
                        "res_obj = stub.%s(empty, timeout=self._request_timeout)"
                        % rpc_method.operation_name,
                    )
                else:
                    self._write(
                        1, "def %s(self, payload):" % rpc_method.method
                    )

                    if status_msg != "":
                        self._write(2, "self.add_warnings('%s')" % status_msg)

                    self._write(2, "pb_obj = json_format.Parse(")
                    self._write(3, "self._serialize_payload(payload),")
                    self._write(3, "pb2.%s()" % rpc_method.request_class)
                    self._write(2, ")")
                    if (
                        self._generate_version_api
                        and rpc_method.method != "get_version"
                    ):
                        self._write(2, "self._do_version_check_once()")
                    "%s=pb_obj" % rpc_method.request_property
                    self._write(
                        2,
                        "req_obj = pb2.{operation_name}Request({request_property}=pb_obj)".format(
                            operation_name=rpc_method.operation_name,
                            request_property=rpc_method.request_property,
                        ),
                    )
                    self._write(2, "stub = self._get_stub()")
                    self._write(
                        2,
                        "res_obj = stub.%s(req_obj, timeout=self._request_timeout)"
                        % rpc_method.operation_name,
                    )
                including_default, return_byte = self._process_good_response(
                    rpc_method
                )
                self._write(2, "response = json_format.MessageToDict(")
                self._write(3, "res_obj, preserving_proto_field_name=True")
                self._write(2, ")")
                self._write(
                    2, 'status_code_200 = response.get("status_code_200")'
                )
                self._write(2, "if status_code_200 is not None:")
                if return_byte:
                    self._write(
                        3, "return io.BytesIO(res_obj.status_code_200)"
                    )
                elif rpc_method.good_response_type:
                    if including_default:
                        self._write(3, "if len(status_code_200) == 0:")
                        self._write(
                            4, "status_code_200 = json_format.MessageToDict("
                        )
                        self._write(5, "res_obj.status_code_200,")
                        self._write(5, "preserving_proto_field_name=True,")
                        self._write(5, "including_default_value_fields=True")
                        self._write(4, ")")
                    self._write(
                        3,
                        "return self.%s().deserialize("
                        % rpc_method.good_response_type,
                    )
                    self._write(4, "status_code_200")
                    self._write(3, ")")
                else:
                    self._write(3, 'return response.get("status_code_200")')
                for rsp_code in rpc_method.bad_responses:
                    self._write(
                        2,
                        """if response.get("status_code_{code}") is not None:""".format(
                            code=rsp_code
                        ),
                    )
                    self._write(
                        3,
                        """raise Exception({code}, response.get("status_code_{code}"))""".format(
                            code=rsp_code
                        ),
                    )

    def _process_good_response(self, rpc_method):
        including_default = False
        return_byte = False
        property = rpc_method.good_response_property
        _, _, _, ref = self._get_object_property_class_names(property)
        if ref is not None:
            property = self._get_object_from_ref(ref)
        content = self._get_parser("$..content").find(property)
        if len(content) > 0 and "application/octet-stream" in content[0].value:
            return_byte = True

        parse_warnings = False
        if len(content) > 0:
            value = content[0].value
            if "application/json" in value:
                schema = value["application/json"].get("schema")
                if schema is not None:
                    _, _, _, ref = self._get_object_property_class_names(
                        schema
                    )
                    if ref is not None:
                        schema = self._get_object_from_ref(ref)
                    if (
                        schema.get("properties") is not None
                        and schema.get("properties").get("warnings")
                        is not None
                    ):
                        parse_warnings = True
        if (
            rpc_method.http_method in ["put", "post", "patch"]
            and parse_warnings
        ):
            including_default = True
        return including_default, return_byte

    def _write_http_api_class(self, methods):
        with open(self._api_filename, "a") as self._fid:
            self._write()
            self._write()
            self._write(0, "class HttpApi(Api):")
            self._write(1, '"""%s' % "OpenAPI HTTP Api")
            self._write(1, '"""')
            self._write(1, "def __init__(self, **kwargs):")
            self._write(2, "super(HttpApi, self).__init__(**kwargs)")
            self._write(2, "self._transport = HttpTransport(**kwargs)")
            self._write()
            self._write(1, "@property")
            self._write(1, "def verify(self):")
            self._write(2, "return self._transport.verify")
            self._write()
            self._write(1, "@verify.setter")
            self._write(1, "def verify(self, value):")
            self._write(2, "self._transport.set_verify(value)")

            for method in methods:
                print("generating method %s" % method["name"])
                self._write()
                if method["x_status"] is not None:
                    key = "{}.{}".format("HttpApi", method["name"])
                    self._deprecated_properties[key] = method["x_status"][1]
                self._write(
                    1,
                    "def %s(%s):"
                    % (method["name"], ", ".join(method["args"])),
                )
                self._write(
                    2,
                    '"""%s %s'
                    % (method["http_method"].upper(), method["url"]),
                )
                self._write(0)
                self._write(2, "%s" % method["description"])
                self._write(0)
                self._write(2, "Return: %s" % method["response_type"])
                self._write(2, '"""')

                if method["x_status"] is not None:
                    status_msg = "%s api is %s, %s" % (
                        method["name"],
                        method["x_status"][0],
                        method["x_status"][1],
                    )
                    self._write(2, "self.add_warnings('%s')" % status_msg)

                if (
                    self._generate_version_api
                    and method["name"] != "get_version"
                ):
                    self._write(2, "self._do_version_check_once()")

                self._write(2, "return self._transport.send_recv(")
                self._write(3, '"%s",' % method["http_method"])
                self._write(3, '"%s",' % method["url"])
                self._write(
                    3,
                    "payload=%s,"
                    % (
                        method["args"][1]
                        if len(method["args"]) > 1
                        else "None"
                    ),
                )
                self._write(
                    3,
                    "return_object=%s,"
                    % (
                        "self." + method["response_type"] + "()"
                        if method["response_type"]
                        else "None"
                    ),
                )
                self._write(2, ")")

    def _write_api_class(self, methods, factories):
        self._generated_classes.append("Api")
        factory_class_name = "Api"
        with open(self._api_filename, "a") as self._fid:
            self._write()
            self._write()
            self._write(0, "class %s(object):" % factory_class_name)
            self._write(1, '"""%s' % "OpenApi Abstract API")
            self._write(1, '"""')
            self._write()
            self._write(1, "__warnings__ = []")
            self._write(1, "def __init__(self, **kwargs):")
            if self._generate_version_api:
                self._write(2, "self._version_meta = self.version()")
                self._write(
                    2,
                    'self._version_meta.api_spec_version = "{}"'.format(
                        self._api_version
                    ),
                )
                self._write(
                    2,
                    'self._version_meta.sdk_version = "{}"'.format(
                        self._sdk_version
                    ),
                )
                self._write(
                    2, 'self._version_check = kwargs.get("version_check")'
                )
                self._write(2, "if self._version_check is None:")
                self._write(3, "self._version_check = False")
                self._write(2, "self._version_check_err = None")
            else:
                self._write(2, "pass")

            self._write()
            self._write(1, "def add_warnings(self, msg):")
            self._write(2, "print('[WARNING]: %s' % msg)")
            self._write(2, "self.__warnings__.append(msg)")

            for method in methods:
                print("generating method %s" % method["name"])
                self._write()
                self._write(
                    1,
                    "def %s(%s):"
                    % (method["name"], ", ".join(method["args"])),
                )
                self._write(
                    2,
                    '"""%s %s'
                    % (method["http_method"].upper(), method["url"]),
                )
                self._write(0)
                self._write(2, "%s" % method["description"])
                self._write(0)
                self._write(2, "Return: %s" % method["response_type"])
                self._write(2, '"""')
                self._write(
                    2, 'raise NotImplementedError("%s")' % method["name"]
                )

            for factory in factories:
                print(
                    "generating top level factory method %s" % factory["name"]
                )
                self._write()
                self._write(1, "def %s(self):" % factory["name"])
                self._write(
                    2,
                    '"""Factory method that creates an instance of %s'
                    % (factory["class_name"]),
                )
                self._write()
                self._write(2, "Return: %s" % factory["class_name"])
                self._write(2, '"""')
                self._write(2, "return %s()" % factory["class_name"])

            self._write()
            self._write(1, "def close(self):")
            self._write(2, "pass")
            # self._write()
            # self._write(1, "def get_api_warnings(self):")
            # self._write(2, "return openapi_warnings")
            # self._write()
            # self._write(1, "def clear_api_warnings(self):")
            # self._write(
            #     2, 'if "2.7" in platform.python_version().rsplit(".", 1)[0]:'
            # )
            # self._write(3, "del openapi_warnings[:]")
            # self._write(2, "else:")
            # self._write(3, "openapi_warnings.clear()")
            self._generate_version_api_methods()

    def _generate_version_api_methods(self):
        if not self._generate_version_api:
            return

        self._write(
            indent=1,
            line="""def _check_client_server_version_compatibility(
        self, client_ver, server_ver, component_name
    ):
        try:
            c = semantic_version.Version(client_ver)
        except Exception as e:
            raise AssertionError(
                "Client {} version '{}' is not a valid semver: {}".format(
                    component_name, client_ver, e
                )
            )

        try:
            s = semantic_version.SimpleSpec(server_ver)
        except Exception as e:
            raise AssertionError(
                "Server {} version '{}' is not a valid semver: {}".format(
                    component_name, server_ver, e
                )
            )


        err = "Client {} version '{}' is not semver compatible with Server {} version '{}'".format(
            component_name, client_ver, component_name, server_ver
        )

        if not s.match(c):
            raise Exception(err)

    def get_local_version(self):
        return self._version_meta

    def get_remote_version(self):
        return self.get_version()

    def check_version_compatibility(self):
        comp_err, api_err = self._do_version_check()
        if comp_err is not None:
            raise comp_err
        if api_err is not None:
            raise api_err

    def _do_version_check(self):
        local = self.get_local_version()
        try:
            remote = self.get_remote_version()
        except Exception as e:
            return None, e

        try:
            self._check_client_server_version_compatibility(
                local.api_spec_version, remote.api_spec_version, "API spec"
            )
        except Exception as e:
            msg = "client SDK version '{}' is not compatible with server SDK version '{}'".format(
                local.sdk_version, remote.sdk_version
            )
            return Exception("{}: {}".format(msg, str(e))), None

        return None, None

    def _do_version_check_once(self):
        if not self._version_check:
            return

        if self._version_check_err is not None:
            raise self._version_check_err

        comp_err, api_err = self._do_version_check()
        if comp_err is not None:
            self._version_check_err = comp_err
            raise comp_err
        if api_err is not None:
            self._version_check_err = None
            raise api_err

        self._version_check = False
        self._version_check_err = None""",
        )

    def _get_object_property_class_names(self, ref):
        """Returns: `Tuple(object_name, property_name, class_name, ref_name)`"""
        object_name = None
        property_name = None
        class_name = None
        ref_name = None
        if isinstance(ref, dict) is True and "$ref" in ref:
            ref_name = ref["$ref"]
        elif isinstance(ref, str) is True:
            ref_name = ref
        if ref_name is not None:
            object_name = ref_name.split("/")[-1]
            property_name = object_name.lower().replace(".", "_")
            class_name = object_name.replace(".", "")
        return (object_name, property_name, class_name, ref_name)

    def _get_external_struct_name(self, openapi_name):
        return self._get_external_field_name(openapi_name).replace("_", "")

    def _get_external_field_name(self, openapi_name):
        """convert openapi fieldname to protobuf fieldname

        - reference: https://developers.google.com/protocol-buffers/docs/reference/go-generated#fields

        Note that the generated Go field names always use camel-case naming,
        even if the field name in the .proto file uses lower-case with underscores (as it should).
        The case-conversion works as follows:
        - The first letter is capitalized for export.
        - NOTE: this is ignored as OpenAPIArt doesn't allow fieldnames to start with an underscore
            - If the first character is an underscore, it is removed and a capital X is prepended.
        - If an interior underscore is followed by a lower-case letter, the underscore is removed, and the following letter is capitalized.
        - NOTE: This isn't documented, if a number is followed by a lower-case letter the following letter is capitalized.
        - Thus, the proto field foo_bar_baz becomes FooBarBaz in Go, and _my_field_name_2 becomes XMyFieldName_2.
        """
        external = ""
        name = openapi_name.replace(".", "")
        for i in range(len(name)):
            if i == 0:
                if name[i] == "_":
                    pass
                else:
                    external += name[i].upper()
            elif name[i] == "_":
                pass
            elif name[i - 1] == "_":
                if name[i].isdigit() or name[i].isupper():
                    external += "_" + name[i]
                else:
                    external += name[i].upper()
            elif name[i - 1].isdigit() and name[i].islower():
                external += name[i].upper()
            else:
                external += name[i]
        if external in ["String"]:
            external += "_"
        return external

    def _write_openapi_object(self, ref, choice_method_name=None):
        schema_object = self._get_object_from_ref(ref)
        ref_name = ref.split("/")[-1]
        class_name = ref_name.replace(".", "")
        if class_name in self._generated_classes:
            return
        self._generated_classes.append(class_name)

        print("generating class %s" % (class_name))
        refs = []
        with open(self._api_filename, "a") as self._fid:
            self._write()
            self._write()
            self._write(0, "class %s(OpenApiObject):" % class_name)
            slots = ["'_parent'"]
            if "choice" in self._get_choice_names(schema_object):
                slots.append("'_choice'")
            self._write(1, "__slots__ = (%s)" % ",".join(slots))
            self._write()

            # write _TYPES definition
            # TODO: this func won't detect whether $ref for a given property is
            # a list because it relies on 'type' attribute to do so
            openapi_types = self._get_openapi_types(schema_object)
            if len(openapi_types) > 0:
                self._write(1, "_TYPES = {")
                for name, value in openapi_types:
                    if len(value) == 1:
                        self._write(
                            2,
                            "'%s': {'%s': %s},"
                            % (
                                name,
                                list(value.keys())[0],
                                list(value.values())[0],
                            ),
                        )
                        continue
                    self._write(2, "'%s': %s" % (name, "{"))
                    for n, v in value.items():
                        if isinstance(v, list):
                            self._write(3, "'%s': [" % n)
                            for i in v:
                                self._write(4, "'%s'," % i)
                            self._write(3, "],")
                            continue
                        self._write(3, "'%s': %s," % (n, v))
                    self._write(2, "},")
                self._write(1, "} # type: Dict[str, str]")
                self._write()
            else:
                # TODO: provide empty types as workaround because deserializer
                # in common.py currently expects it
                self._write(1, "_TYPES = {} # type: Dict[str, str]")
                self._write()

            required, defaults = self._get_required_and_defaults(schema_object)

            if len(required) > 0:
                self._write(
                    1, "_REQUIRED = {} # type: tuple(str)".format(required)
                )
                self._write()
            else:
                self._write(1, "_REQUIRED= () # type: tuple(str)")
                self._write()

            if len(defaults) > 0:
                self._write(1, "_DEFAULTS = {")
                for name, value in defaults:
                    if isinstance(value, (list, bool, int, float, tuple)):
                        self._write(2, "'%s': %s," % (name, value))
                    else:
                        self._write(2, "'%s': '%s'," % (name, value))
                self._write(1, "} # type: Dict[str, Union(type)]")
                self._write()
            else:
                self._write(1, "_DEFAULTS= {} # type: Dict[str, Union(type)]")
                self._write()

            # write constants
            # search for all simple properties with enum or
            # x-constant and add them here
            for enum in self._get_parser("$..enum | x-constants").find(
                schema_object
            ):
                for name in enum.value:
                    value = name
                    value_type = "string"
                    if isinstance(enum.value, dict):
                        value = enum.value[name]
                        value_type = (
                            enum.context.value["type"]
                            if "type" in enum.context.value
                            else "string"
                        )
                    if value_type == "string":
                        self._write(
                            1, "%s = '%s' # type: str" % (name.upper(), value)
                        )
                    else:
                        self._write(1, "%s = %s #" % (name.upper(), value))
                if len(enum.value) > 0:
                    self._write()

            # find x-status codes
            status = self._get_status_dict(schema_object, class_name)
            if len(status) > 0:
                self._write(1, "_STATUS = {")
                for name, value in status.items():
                    if isinstance(name, (list, bool, int, float, tuple)):
                        self._write(2, "%s: '%s'," % (name, value))
                    else:
                        self._write(2, "'%s': '%s'," % (name, value))
                self._write(1, "} # type: Dict[str, Union(type)]")
                self._write()
            else:
                self._write(1, "_STATUS= {} # type: Dict[str, Union(type)]")
                self._write()

            # write def __init__(self)
            params = "self, parent=None"
            if "choice" in self._get_choice_names(schema_object):
                params += ", choice=None"
            init_params, properties, _ = self._get_property_param_string(
                schema_object
            )
            params = (
                params
                if len(init_params) == 0
                else ", ".join([params, init_params])
            )
            self._write(1, "def __init__(%s):" % (params))
            self._write(2, "super(%s, self).__init__()" % class_name)
            self._write(2, "self._parent = parent")
            for property_name in properties:
                self._write(
                    2,
                    "self._set_property('%s', %s)"
                    % (property_name, property_name),
                )
            if "choice" in self._get_choice_names(schema_object):
                self._write(
                    2,
                    "if 'choice' in self._DEFAULTS and choice is None and self._DEFAULTS['choice'] in self._TYPES:",
                )
                self._write(3, "getattr(self, self._DEFAULTS['choice'])")
                self._write(2, "else:")
                self._write(3, "self._set_property('choice', choice)")

            # write def set(self)
            self._write_set_method(schema_object)

            # process properties - TBD use this one level up to process
            # schema, in requestBody, Response and also
            refs = self._process_properties(
                class_name,
                schema_object,
                choice_child=choice_method_name is not None,
            )

        # descend into child properties
        for ref in refs:
            self._write_openapi_object(ref[0], ref[3])
            if ref[1] is True:
                self._write_openapi_list(ref[0], ref[2])

    def _write_set_method(self, schema_object):
        write_set = False
        if "choice" in self._get_choice_names(schema_object):
            write_set = False
        init_params, properties, _ = self._get_property_param_string(
            schema_object
        )
        if len(init_params) > 0:
            write_set = True
        params = ["self"]
        for property in properties:
            str = property + "=None"
            params.append(str)
        params = params if len(init_params) == 0 else ", ".join(params)
        if write_set:
            self._write(1, "def set(%s):" % (params))
            self._write(
                2, "for property_name, property_value in locals().items():"
            )
            self._write(
                3,
                "if property_name != 'self' and property_value is not None:",
            )
            self._write(4, "self._set_property(property_name, property_value)")

    def _get_simple_type_names(self, schema_object):
        simple_type_names = []
        if "properties" in schema_object:
            choice_names = self._get_choice_names(schema_object)
            for name in schema_object["properties"]:
                if name in choice_names:
                    continue
                ref = self._get_parser("$..'$ref'").find(
                    schema_object["properties"][name]
                )
                if len(ref) == 0:
                    simple_type_names.append(name)
        return simple_type_names

    def _get_choice_names(self, schema_object):
        choice_names = []
        if (
            "properties" in schema_object
            and "choice" in schema_object["properties"]
        ):
            choice_names = schema_object["properties"]["choice"]["enum"][:]
            choice_names.append("choice")
        return choice_names

    def _process_properties(
        self, class_name=None, schema_object=None, choice_child=False
    ):
        """Process all properties of a /component/schema object
        Write a factory method for all choice
        If there are no properties then the schema_object is a primitive or array type
        """
        refs = []
        if "properties" in schema_object:
            choice_names = self._get_choice_names(schema_object)
            excluded_property_names = []
            for choice_name in choice_names:

                # this code is to allow choices with no properties
                if choice_name not in schema_object["properties"]:
                    excluded_property_names.append(choice_name)
                    continue

                if "$ref" not in schema_object["properties"][choice_name]:
                    continue
                ref = schema_object["properties"][choice_name]["$ref"]
                self._write_factory_method(
                    None, choice_name, ref, property_status=None
                )
                excluded_property_names.append(choice_name)
            for property_name in schema_object["properties"]:
                if property_name in excluded_property_names:
                    continue
                property = schema_object["properties"][property_name]
                write_set_choice = (
                    property_name in choice_names and property_name != "choice"
                )
                self._write_openapi_property(
                    schema_object,
                    property_name,
                    property,
                    class_name,
                    write_set_choice,
                )
            for property_name, property in schema_object["properties"].items():
                ref = self._get_parser("$..'$ref'").find(property)
                if len(ref) > 0:
                    restriction = self._get_type_restriction(property)
                    choice_name = (
                        property_name
                        if property_name in excluded_property_names
                        else None
                    )
                    refs.append(
                        (
                            ref[0].value,
                            restriction.startswith("List["),
                            property_name,
                            choice_name,
                        )
                    )
        return refs

    def _write_openapi_list(self, ref, property_name):
        """This is the class writer for schema object properties that are of
        type array with a ref to an object.  The class should provide a factory
        method for the encapsulated ref.
        ```
        properties:
          ports:
            type: array
            items:
              $ref: '#/components/schema/...'
        ```

        If the schema object has a property named choice, that property needs
        to be brought forward so that the generated class can provide factory
        methods for objects for each of the choice $refs (if any).

        if choice exists:
            for each choice enum that is a $ref:
                generate a factory method named after the choice
                in the method set the choice property
        """
        yobject = self._get_object_from_ref(ref)
        ref_name = ref.split("/")[-1]
        contained_class_name = ref_name.replace(".", "")
        class_name = "%sIter" % contained_class_name
        if class_name in self._generated_classes:
            return
        self._generated_classes.append(class_name)

        self._imports = []
        print("generating class %s" % (class_name))
        with open(self._api_filename, "a") as self._fid:
            self._write()
            self._write()
            self._write(0, "class %s(OpenApiIter):" % class_name)
            self._write(1, "__slots__ = ('_parent', '_choice')")
            self._write()

            # if all choice(s) are $ref, the getitem should return the actual choice object
            # the _GETITEM_RETURNS_CHOICE_OBJECT class static allows the OpenApiIter to
            # correctly return the selected choice if any
            get_item_returns_choice = True
            if "properties" in yobject and "choice" in yobject["properties"]:
                for property, item in yobject["properties"].items():
                    if property == "choice":
                        continue
                    elif (
                        property in yobject["properties"]["choice"]["enum"]
                        and len(self._get_parser('$.."$ref"').find(item)) > 0
                    ):
                        continue
                    get_item_returns_choice = False
                    break
            else:
                get_item_returns_choice = False
            self._write(
                1,
                "_GETITEM_RETURNS_CHOICE_OBJECT = {}".format(
                    get_item_returns_choice
                ),
            )

            self._write()
            self._write(1, "def __init__(self, parent=None, choice=None):")
            self._write(2, "super(%s, self).__init__()" % class_name)
            self._write(2, "self._parent = parent")
            self._write(2, "self._choice = choice")

            # write container emulation methods __getitem__, __iter__, __next__
            self._write_openapilist_special_methods(
                contained_class_name, yobject
            )

            # write a factory method for the schema object in the list that returns the container
            self._write_factory_method(
                contained_class_name,
                ref_name.lower().split(".")[-1],
                ref,
                True,
                False,
            )

            # write an add method for the schema object in the list that creates and returns the new object
            self._write_add_method(
                yobject,
                ref,
                False,
                class_name,
                contained_class_name,
                class_name,
            )

            # write choice factory methods if the only properties are choice properties
            if get_item_returns_choice is True:
                for property, item in yobject["properties"].items():
                    if property == "choice":
                        continue
                    self._write_factory_method(
                        contained_class_name,
                        property,
                        item["$ref"],
                        True,
                        True,
                    )

        return class_name

    def _write_openapilist_special_methods(
        self, contained_class_name, schema_object
    ):
        get_item_class_names = [contained_class_name]
        if (
            "properties" in schema_object
            and "choice" in schema_object["properties"]
        ):
            for property in schema_object["properties"]:
                if property in schema_object["properties"]["choice"]["enum"]:
                    if "$ref" in schema_object["properties"][property]:
                        ref = schema_object["properties"][property]["$ref"]
                        (
                            _,
                            _,
                            choice_class_name,
                            _,
                        ) = self._get_object_property_class_names(ref)
                        if choice_class_name not in get_item_class_names:
                            get_item_class_names.append(choice_class_name)
        get_item_class_names.sort()
        self._write()
        self._write(1, "def __getitem__(self, key):")
        self._write(
            2, "# type: (str) -> Union[%s]" % (", ".join(get_item_class_names))
        )
        self._write(2, "return self._getitem(key)")
        self._write()
        self._write(1, "def __iter__(self):")
        self._write(2, "# type: () -> %sIter" % contained_class_name)
        self._write(2, "return self._iter()")
        self._write()
        self._write(1, "def __next__(self):")
        self._write(2, "# type: () -> %s" % contained_class_name)
        self._write(2, "return self._next()")
        self._write()
        self._write(1, "def next(self):")
        self._write(2, "# type: () -> %s" % contained_class_name)
        self._write(2, "return self._next()")
        self._write()
        self._write(1, "def _instanceOf(self, item):")
        self._write(2, "if not isinstance(item, %s):" % (contained_class_name))
        self._write(
            3,
            'raise Exception("Item is not an instance of %s")'
            % (contained_class_name),
        )

    def _write_factory_method(
        self,
        contained_class_name,
        method_name,
        ref,
        openapi_list=False,
        choice_method=False,
        property_status=None,
    ):
        yobject = self._get_object_from_ref(ref)
        _, _, class_name, _ = self._get_object_property_class_names(ref)
        (
            param_string,
            properties,
            type_string,
        ) = self._get_property_param_string(yobject)
        self._write()
        if openapi_list is True:
            self._imports.append(
                "from .%s import %s" % (class_name.lower(), class_name)
            )
            params = (
                "self"
                if len(param_string) == 0
                else ", ".join(["self", param_string])
            )
            self._write(1, "def %s(%s):" % (method_name, params))
            return_class_name = class_name
            if contained_class_name is not None:
                return_class_name = "{}Iter".format(contained_class_name)
            self._write(
                2, "# type: (%s) -> %s" % (type_string, return_class_name)
            )
            self._write(
                2,
                '"""Factory method that creates an instance of the %s class'
                % (class_name),
            )
            self._write()
            self._write(2, "%s" % self._get_description(yobject))
            self._write()
            self._write(2, "Returns: %s" % (return_class_name))
            self._write(2, '"""')
            if choice_method is True:
                self._write(2, "item = %s()" % (contained_class_name))
                self._write(2, "item.%s" % (method_name))
                self._write(2, "item.choice = '%s'" % (method_name))
            else:
                params = ["parent=self._parent"]
                if (
                    "properties" in yobject
                    and "choice" in yobject["properties"]
                ):
                    params.append("choice=self._choice")
                params.extend(["%s=%s" % (name, name) for name in properties])
                self._write(
                    2, "item = %s(%s)" % (class_name, ", ".join(params))
                )
            self._write(2, "self._add(item)")
            self._write(2, "return self")
            self._write()
        else:
            self._write(1, "@property")
            self._write(1, "def %s(self):" % (method_name))
            self._write(2, "# type: () -> %s" % (class_name))
            self._write(
                2,
                '"""Factory property that returns an instance of the %s class'
                % (class_name),
            )
            self._write()
            self._write(2, "%s" % self._get_description(yobject))
            self._write()
            self._write(2, "Returns: %s" % (class_name))
            self._write(2, '"""')
            self._write(
                2,
                "return self._get_property('%s', %s, self, '%s')"
                % (method_name, class_name, method_name),
            )

    def _write_add_method(
        self,
        yobject,
        ref,
        choice_method,
        class_name,
        contained_class_name,
        return_class_name,
    ):
        """Writes an add method"""
        method_name = ref.lower().split("/")[-1]
        self._imports.append(
            "from .%s import %s"
            % (contained_class_name.lower(), contained_class_name)
        )
        (
            param_string,
            properties,
            type_string,
        ) = self._get_property_param_string(yobject)
        params = (
            "self"
            if len(param_string) == 0
            else ", ".join(["self", param_string])
        )
        self._write(1, "def add(%s):" % (params))
        self._write(
            2, "# type: (%s) -> %s" % (type_string, contained_class_name)
        )
        self._write(
            2,
            '"""Add method that creates and returns an instance of the %s class'
            % (contained_class_name),
        )
        self._write()
        self._write(2, "%s" % self._get_description(yobject))
        self._write()
        self._write(2, "Returns: %s" % (contained_class_name))
        self._write(2, '"""')
        if choice_method is True:
            self._write(2, "item = self.%s()" % (method_name))
            self._write(2, "item.%s" % (contained_class_name))
            self._write(2, "item.choice = '%s'" % (contained_class_name))
        else:
            params = ["parent=self._parent"]
            if "properties" in yobject and "choice" in yobject["properties"]:
                params.append("choice=self._choice")
            params.extend(["%s=%s" % (name, name) for name in properties])
            self._write(
                2, "item = %s(%s)" % (contained_class_name, ", ".join(params))
            )
        self._write(2, "self._add(item)")
        self._write(2, "return item")

    def _get_property_param_string(self, yobject):
        property_param_string = []
        property_type_string = []
        properties = []
        if "properties" in yobject:
            for name, property in yobject["properties"].items():
                if name == "choice":
                    continue
                default = None
                type_string = self._get_type_restriction(property)
                if "obj" not in type_string:
                    if "default" in property:
                        default = property["default"]
                    if name == "choice":
                        val = "None"
                    elif property["type"] in [
                        "number",
                        "integer",
                        "boolean",
                        "array",
                    ]:
                        val = "None" if default is None else default
                    else:
                        val = (
                            "None"
                            if default is None
                            else "'{}'".format(default.strip())
                        )
                    properties.append(name)
                    property_param_string.append("%s=%s" % (name, val))
                    property_type_string.append(type_string)
        types = ",".join(property_type_string)
        return (", ".join(property_param_string), properties, types)

    def _write_openapi_property(
        self, schema_object, name, property, klass_name, write_set_choice=False
    ):
        ref = self._get_parser("$..'$ref'").find(property)
        restriction = self._get_type_restriction(property)
        if len(ref) > 0:
            object_name = ref[0].value.split("/")[-1]
            class_name = object_name.replace(".", "")
            if restriction.startswith("List["):
                type_name = "%sIter" % class_name
            else:
                type_name = class_name
        else:
            type_name = restriction
        self._write()
        self._write(1, "@property")
        self._write(1, "def %s(self):" % name)
        self._write(2, "# type: () -> %s" % (type_name))
        self._write(2, '"""%s getter' % (name))
        self._write()
        self._write(2, self._get_description(property))
        self._write()
        self._write(2, "Returns: %s" % type_name)
        self._write(2, '"""')
        if (
            len(self._get_parser("$..'type'").find(property)) > 0
            and len(ref) == 0
        ):
            self._write(2, "return self._get_property('%s')" % (name))
            self._write()
            if name == "auto":
                return
            self._write(1, "@%s.setter" % name)
            self._write(1, "def %s(self, value):" % name)
            self._write(2, '"""%s setter' % (name))
            self._write()
            self._write(2, self._get_description(property))
            self._write()
            self._write(2, "value: %s" % restriction)
            self._write(2, '"""')
            required, defaults = self._get_required_and_defaults(schema_object)
            if len(required) > 0:
                if name in required:
                    self._write(2, "if value is None:")
                    self._write(
                        3,
                        "raise TypeError('Cannot set required property %s as None')"
                        % name,
                    )
            if write_set_choice is True:
                self._write(
                    2, "self._set_property('%s', value, '%s')" % (name, name)
                )
            else:
                self._write(2, "self._set_property('%s', value)" % (name))
        elif len(ref) > 0:
            if restriction.startswith("List["):
                self._write(
                    2,
                    "return self._get_property('%s', %sIter, self._parent, self._choice)"
                    % (name, class_name),
                )
            else:
                self._write(
                    2,
                    "return self._get_property('%s', %s)" % (name, class_name),
                )

    def _get_description(self, yobject):
        if "description" not in yobject:
            yobject["description"] = "TBD"
        # remove tabs, multiple spaces
        description = re.sub(r"\n", ". ", yobject["description"])
        description = re.sub(r"\s+", " ", description)
        description = re.sub(r" . ", " ", description)
        return description
        # doc_string = []
        # for line in re.split('\. ', description):
        #     line = re.sub('\.$', '', line)
        #     if len(line) > 0:
        #         doc_string.append('%s  ' % line)
        # return doc_string

    def _get_status_dict(self, yobject, class_name):
        status = {}

        if yobject.get("x-status", None) is not None:
            status["self"] = self._get_status_msg(class_name, yobject, "class")

        if yobject.get("properties", None) is None:
            return status

        for name in yobject["properties"]:
            property_value = yobject["properties"][name]
            if "x-status" in property_value:
                status[name] = self._get_status_msg(
                    name, property_value, "prop", class_name
                )
            elif "x-enum" in property_value:
                for enum in property_value["x-enum"]:
                    enum_value = property_value["x-enum"][enum]
                    if "x-status" in enum_value:
                        key = "%s.%s" % (name, enum)
                        status[key] = self._get_status_msg(
                            enum.upper(), enum_value, "enum", name
                        )

        return status

    def _get_status_msg(
        self, property_name, x_status_obj, prop_type, parent_name=None
    ):
        status_type = x_status_obj["x-status"].get("status")
        info = x_status_obj["x-status"].get("information")

        status_msg = ""
        if prop_type == "prop":
            status_msg = "%s property in schema %s is %s, %s" % (
                property_name,
                parent_name,
                status_type,
                info,
            )
            pass
        elif prop_type == "enum":
            status_msg = "%s enum in property %s is %s, %s" % (
                property_name,
                parent_name,
                status_type,
                info,
            )
        else:
            status_msg = "%s is %s, %s" % (property_name, status_type, info)

        return status_msg

    def _get_data_types(self, yproperty):
        data_type_map = {
            "integer": "int",
            "string": "str",
            "boolean": "bool",
            "array": "list",
            "number": "float",
            "float": "float",
            "double": "float",
        }
        if yproperty["type"] in data_type_map:
            return data_type_map[yproperty["type"]]
        else:
            return yproperty["type"]

    def _get_openapi_types(self, yobject):
        types = []
        if "properties" in yobject:
            for name in yobject["properties"]:
                yproperty = yobject["properties"][name]
                ref = self._get_parser("$..'$ref'").find(yproperty)
                pt = {}
                if "type" in yproperty:
                    pt.update({"type": self._get_data_types(yproperty)})
                    pt.update(
                        {"enum": yproperty["enum"]}
                    ) if "enum" in yproperty else None
                    pt.update(
                        {"format": "'%s'" % yproperty["format"]}
                    ) if "format" in yproperty else None
                if len(ref) > 0:
                    object_name = ref[0].value.split("/")[-1]
                    class_name = object_name.replace(".", "")
                    if "type" in yproperty and yproperty["type"] == "array":
                        class_name += "Iter"
                    pt.update({"type": "'%s'" % class_name})
                if (
                    len(ref) == 0
                    and "items" in yproperty
                    and "type" in yproperty["items"]
                ):
                    pt.update(
                        {"itemtype": self._get_data_types(yproperty["items"])}
                    )
                    if "format" in yproperty["items"]:
                        pt.update(
                            {
                                "itemformat": "'%s'"
                                % yproperty["items"]["format"]
                            }
                        )
                min_max = yproperty.get("maximum", yproperty.get("minimum", 0))
                key = (
                    "itemformat"
                    if pt.get("itemtype") is not None
                    else "format"
                )
                if min_max > 2147483647:
                    pt.update({key: r"'int64'"})
                if len(ref) == 0 and "minimum" in yproperty:
                    pt.update({"minimum": yproperty["minimum"]})
                if len(ref) == 0 and "maximum" in yproperty:
                    pt.update({"maximum": yproperty["maximum"]})
                if len(ref) == 0 and "minLength" in yproperty:
                    pt.update({"minLength": yproperty["minLength"]})
                if len(ref) == 0 and "maxLength" in yproperty:
                    pt.update({"maxLength": yproperty["maxLength"]})
                if len(pt) > 0:
                    types.append((name, pt))
                # TODO: restore behavior
                # if "x-constraint" in yproperty:
                #     cons_lst = []
                #     for cons in yproperty["x-constraint"]:
                #         ref, prop = cons.split("/properties/")
                #         klass = self._get_classname_from_ref(ref)
                #         cons_lst.append("%s.%s" % (klass, prop.strip("/")))
                #     if cons_lst != []:
                #         pt.update({"constraint": cons_lst})
                # if "x-unique" in yproperty:
                #     pt.update({"unique": '"%s"' % yproperty["x-unique"]})
        return types

    def _get_required_and_defaults(self, yobject):
        required = []
        defaults = []
        if "required" in yobject:
            required = yobject["required"]
        if "properties" in yobject:
            for name in yobject["properties"]:
                yproperty = yobject["properties"][name]
                if "default" in yproperty:
                    default = yproperty["default"]
                    if "type" in yproperty and yproperty["type"] == "number":
                        default = float(default)
                    defaults.append((name, default))
        return (tuple(required), defaults)

    def _get_default_value(self, property):
        if "default" in property:
            return property["default"]
        property_type = property["type"]
        if property_type == "array":
            return "[]"
        if property_type == "string":
            return "''"
        if property_type == "integer":
            return 0
        if property_type == "number":
            return 0
        if property_type == "bool":
            return False
        raise Exception(
            "Missing handler for property type `%s`" % property_type
        )

    def _get_api_paths(self):
        paths = []
        for url, yobject in self._openapi["paths"].items():
            for method in yobject:
                if method.lower() in ["get", "post", "put", "patch", "delete"]:
                    paths.append(
                        {
                            "url": url,
                            "method": method,
                            "operation": yobject[method],
                        }
                    )
        return paths

    def _write_data_properties(self, schema, classname, choice_tuples):
        if len(choice_tuples) > 0:
            choices = []
            for choice_tuple in choice_tuples:
                choices.append(choice_tuple[0])
            self._write(
                2,
                "if isinstance(choice, (%s)) is False:" % (", ".join(choices)),
            )
            self._write(
                3,
                "raise TypeError('choice must be of type: %s')"
                % (", ".join(choices)),
            )
            self._write(
                2,
                "self.__setattr__('choice', %s._CHOICE_MAP[type(choice).__name__])"
                % classname,
            )
            self._write(
                2,
                "self.__setattr__(%s._CHOICE_MAP[type(choice).__name__], choice)"
                % classname,
            )

        if "properties" in schema:
            for name, property in schema["properties"].items():
                if (
                    len([item for item in choice_tuples if item[1] == name])
                    == 0
                    and name != "choice"
                ):
                    restriction = self._get_isinstance_restriction(
                        schema, name, property
                    )
                    self._write(
                        2,
                        "if isinstance(%s, %s) is True:" % (name, restriction),
                    )
                    if restriction == "(list, type(None))":
                        self._write(
                            3,
                            "self.%s = [] if %s is None else list(%s)"
                            % (name, name, name),
                        )
                    else:
                        if "pattern" in property:
                            self._write(3, "import re")
                            self._write(
                                3,
                                "assert(bool(re.match(r'%s', %s)) is True)"
                                % (property["pattern"], name),
                            )
                        self._write(3, "self.%s = %s" % (name, name))
                    self._write(2, "else:")
                    self._write(
                        3,
                        "raise TypeError('%s must be an instance of %s')"
                        % (name, restriction),
                    )

    def _get_isinstance_restriction(self, schema, name, property):
        type_none = ", type(None)"
        if "required" in schema and name in schema["required"]:
            type_none = ""
        if "$ref" in property:
            return "(%s%s)" % (
                self._get_classname_from_ref(property["$ref"]),
                type_none,
            )
        elif name == "additionalProperties":
            return "**additional_properties"
        elif property["type"] in ["number", "integer"]:
            return "(float, int%s)" % type_none
        elif property["type"] == "string":
            return "(str%s)" % type_none
        elif property["type"] == "array":
            return "(list%s)" % type_none
        elif property["type"] == "boolean":
            return "(bool%s)" % type_none

    def _get_type_restriction(self, property):
        try:
            if "$ref" in property:
                ref_obj = self._get_object_from_ref(property["$ref"])
                description = ""
                if "description" in ref_obj:
                    description = ref_obj["description"]
                if "description" in property:
                    description += property["description"]
                property["description"] = description
                class_name = property["$ref"].split("/")[-1].replace(".", "")
                return "obj(%s)" % class_name
            elif "oneOf" in property:
                return "Union[%s]" % ",".join(
                    [item["type"] for item in property["oneOf"]]
                )
            elif property["type"] == "number":
                return "float"
            elif property["type"] == "integer":
                return "int"
            elif property["type"] == "string":
                if "enum" in property:
                    values = property["enum"]
                    values.sort()
                    values = ['Literal["{}"]'.format(s) for s in values]
                    return "Union[%s]" % ", ".join(values)
                else:
                    return "str"
            elif property["type"] == "array":
                return "List[%s]" % self._get_type_restriction(
                    property["items"]
                )
            elif property["type"] == "boolean":
                return "bool"
        except Exception as e:
            print("Error ", property, e)
            raise e

    def _get_object_from_ref(self, ref):
        leaf = self._openapi
        for attr in ref.split("/")[1:]:
            leaf = leaf[attr]
        return leaf

    def _get_classname_from_ref(self, ref):
        final_piece = ref.split("/")[-1]
        if "." in final_piece:
            return final_piece.split(".")[-1]
        else:
            return final_piece

    def _write(self, indent=0, line=""):
        self._fid.write("    " * indent + line + "\n")



import pytest


def test_defaults_and_stale_choices(api):
    config = api.prefix_config()
    config.f
    f = {"choice": "f_a", "f_a": "some string"}
    assert config.f.serialize(config.DICT) == f

    f = {"choice": "f_b", "f_b": 4.25}
    config = api.prefix_config()
    config.f.deserialize(f)
    assert config.f._properties == f
    assert config.f.serialize(config.DICT) == f


def test_required(api):
    config = api.prefix_config()
    try:
        config.mandatory
        config.mandatory.serialize()
        pytest.fail("config got validated")
    except ValueError:
        pass


def test_serialize_deserialize(api):
    config = api.prefix_config()
    config.a = "asdf"
    config.b = 1.1
    config.c = 1
    config.required_object.e_a = 1.1
    config.required_object.e_b = 1.2
    config.d_values = [config.A, config.B, config.C]
    config.level.l1_p1.l2_p1.l3_p1 = "test"
    config.level.l1_p2.l4_p1.l1_p2.l4_p1.l1_p1.l2_p1.l3_p1 = "test"
    api.set_config(config)
    _config = api.get_config()
    assert config.serialize(config.DICT) == _config.serialize(_config.DICT)


def test_x_include(api):
    config = api.prefix_config()
    config.a = "asdf"
    config.b = 1.1
    config.c = 1
    config.e.e_a = 0.1
    config.e.e_b = 0.2
    assert "name" in config.e._TYPES
    assert "m_param1" in config.e._TYPES
    assert "m_param2" in config.e._TYPES


@pytest.mark.parametrize("ipv4", ["1.1.1.1", "01.002.003.4", "0.0.0.0"])
def test_x_pattern_ipv4_good_value(default_config, ipv4):
    default_config.ipv4_pattern.ipv4.value = ipv4
    try:
        default_config.ipv4_pattern.serialize(default_config.DICT)
    except Exception:
        pytest.fail("ipv4 pattern failed to serialize")


@pytest.mark.parametrize(
    "ipv6",
    [
        "::",
        " :: ",
        "1::",
        "::1",
        "abcd::abcd",
        "abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd",
    ],
)
def test_x_pattern_ipv6_good_value(default_config, ipv6):
    default_config.ipv6_pattern.ipv6.value = ipv6
    try:
        default_config.ipv6_pattern.serialize(default_config.DICT)
    except Exception:
        pytest.fail("ipv6 pattern failed to serialize")


@pytest.mark.parametrize("mac", ["ab:ab:ab:ab:ab:ab"])
def test_x_pattern_mac_good_value(default_config, mac):
    default_config.mac_pattern.mac.value = mac
    try:
        default_config.mac_pattern.serialize(default_config.DICT)
    except Exception:
        pytest.fail("mac pattern failed to serialize")


@pytest.mark.parametrize("integer", [0, 100, 255])
def test_x_pattern_integer_good_value(default_config, integer):
    default_config.integer_pattern.integer.value = integer
    try:
        default_config.integer_pattern.serialize(default_config.DICT)
    except Exception:
        pytest.fail("integer pattern failed to serialize")


@pytest.mark.parametrize("ipv4", [["1.1.1.1", "-255.-255.-255.-255"]])
def test_x_pattern_ipv4_good_and_bad_list(default_config, ipv4):
    default_config.ipv4_pattern.ipv4.values = ipv4
    try:
        default_config.ipv4_pattern.serialize(default_config.DICT)
        pytest.fail("ipv4 values got serialize")
    except TypeError as e:
        if "['-255.-255.-255.-255']" not in str(e):
            pytest.fail("Invalid ipv4 list is not proper in error message")


@pytest.mark.parametrize("ipv6", [["::", ":", "abcd::abcd::"]])
def test_x_pattern_ipv6_good_and_bad_list(default_config, ipv6):
    default_config.ipv6_pattern.ipv6.values = ipv6
    try:
        default_config.ipv6_pattern.serialize(default_config.DICT)
        pytest.fail("ipv6 values got serialize")
    except TypeError as e:
        if "[':', 'abcd::abcd::']" not in str(e):
            pytest.fail("Invalid ipv6 list is not proper in error message")


@pytest.mark.parametrize("mac", [[":", "abcd::abcd::", "ab:ab:ab:ab:ab:ab"]])
def test_x_pattern_mac_good_and_bad_list(default_config, mac):
    default_config.mac_pattern.mac.values = mac
    try:
        default_config.mac_pattern.serialize(default_config.DICT)
        pytest.fail("mac values got serialize")
    except TypeError as e:
        if "[':', 'abcd::abcd::']" not in str(e):
            pytest.fail("Invalid mac list is not proper in error message")


@pytest.mark.parametrize(
    "integer", [["abcd::abcd::", 10, 3, 255, 256, "ab:ab:ab:ab:ab:ab"]]
)
def test_x_pattern_integer_good_and_bad_list(default_config, integer):
    default_config.integer_pattern.integer.values = integer
    try:
        default_config.integer_pattern.serialize(default_config.DICT)
        pytest.fail("integer values got serialize")
    except TypeError as e:
        if "['abcd::abcd::', 256, 'ab:ab:ab:ab:ab:ab']" not in str(e):
            pytest.fail("Invalid integer list is not proper in error message")


@pytest.mark.parametrize("direction", ["increment", "decrement"])
@pytest.mark.parametrize("index", list(range(4)))
def test_x_pattern_good_inc_dec(default_config, index, direction):
    start = ["1.1.1.1", "2000::1", "00:00:00:00:00:0a", 10]
    step = ["0.1.1.0", "::1", "00:00:00:00:00:01", 1]
    count = [100, 100, 100, 100]
    enum = ["ipv4", "ipv6", "mac", "integer"]
    enum_obj = getattr(
        getattr(default_config, "%s_pattern" % enum[index]), enum[index]
    )
    dir_obj = getattr(enum_obj, direction)
    dir_obj.start = start[index]
    dir_obj.step = step[index]
    dir_obj.count = count[index]
    try:
        default_config.serialize(default_config.DICT)
    except TypeError:
        pytest.fail("%s with %s Failed to serialize" % (enum, direction))


@pytest.mark.parametrize("direction", ["increment", "decrement"])
@pytest.mark.parametrize("index", list(range(4)))
def test_x_pattern_bad_inc_dec(default_config, index, direction):
    start = ["1.1.1.1", "2000::1", "00:00:00:00:00:00", 10]
    step = [1, 10, "00:00:00:00:0001", -1]
    count = [100, 100, 100, 100]
    enum = ["ipv4", "ipv6", "mac", "integer"]
    enum_obj = getattr(
        getattr(default_config, "%s_pattern" % enum[index]), enum[index]
    )
    dir_obj = getattr(enum_obj, direction)
    dir_obj.start = start[index]
    dir_obj.step = step[index]
    dir_obj.count = count[index]
    try:
        default_config.serialize(default_config.DICT)
        pytest.fail("%s with %s got serialized" % (enum, direction))
    except TypeError as e:
        print(e)


def test_int_64_format(api, default_config):
    default_config.integer64 = 1000
    dt = default_config.serialize(default_config.DICT)
    assert isinstance(dt["integer64"], str)
    conf = api.prefix_config()
    conf.deserialize(dt)
    assert isinstance(conf.integer64, int)
    conf.integer64 = 2000
    conf.validate()
    conf.integer64 = "2000"
    try:
        conf.validate()
    except TypeError as e:
        print(e)


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])
